let log = Logs.Src.create "custom_options" ~doc:"IP option management"

module Log = (val Logs.src_log log : Logs.LOG)

type entry_type = { id_val : int; inter_val : int; cmd_val : int }

(* Define a record to represent the Custom Option *)
type t = {
  option_type : int; (* 1 byte *)
  length : int; (* 1 byte: total length of the option *)
  count : int; (* 2 byte: number of entries *)
  entries : entry_type list; (* List of entries *)
}

let entry_size = 4

let print_entry_type entry =
  Logs.err (fun m -> m "{ id_val = %d; inter_val = %d; cmd_val = %d }" entry.id_val entry.inter_val entry.cmd_val)

(* Function to print a value of type t *)
let print_t t =
  Logs.err (fun m -> m "\n");
  Logs.err (fun m -> m "  option_type = %d;" t.option_type);
  Logs.err (fun m -> m "  length = %d;" t.length);
  Logs.err (fun m -> m "  count = %d;" t.count);
  Logs.err (fun m -> m "  entries = [");
  List.iter print_entry_type t.entries;
  Logs.err (fun m -> m "  ];");
  Logs.err (fun m -> m "}\n")

(* Create a map module with IP addresses as keys and value tuples as values *)
module IpV4Map = Map.Make (Ipaddr.V4)

let map_entry_to_string ip (id_val, inter_val) =
  Printf.sprintf "%s -> (id_val = %ld, inter_val = %d)" (Ipaddr.V4.to_string ip) id_val inter_val

(* Function to convert the entire map to a string *)
let map_to_string map =
  let entries = IpV4Map.bindings map in
  let entry_strings = List.map (fun (ip, value) -> map_entry_to_string ip value) entries in
  "{" ^ String.concat "; " entry_strings ^ "}"

(* Function to log the map using Logs.err *)
let log_map map =
  let map_str = map_to_string map in
  Log.err (fun m -> m "%s" map_str)

let cmd_DECISION_DROP = 0
let cmd_REGISTER = 1
let map_ip = ref IpV4Map.empty

let add_to_map ip id_val inter_val map =
  let value = (id_val, inter_val) in
  IpV4Map.add ip value !map

let create id_val inter_val cmd_val = { option_type = 31; length = 4 + entry_size; count = 1; entries = [ { id_val; inter_val; cmd_val } ] }
let init () = { option_type = 31; length = entry_size; count = 0; entries = [] }

let add_entry t id_val inter_val cmd_val =
  { option_type = 31; length = t.length + entry_size; count = t.count + 1; entries = { id_val; inter_val; cmd_val } :: t.entries }

let of_cstruct buf =
  let option_type = Cstruct.get_uint8 buf 0 in
  let length = Cstruct.get_uint8 buf 1 in
  let count = Cstruct.BE.get_uint16 buf 2 in
  let rec read_entries buf count lst offset =
    match count with
    | 0 -> lst
    | _ ->
        let id_val = Cstruct.BE.get_uint16 buf offset in
        let inter_val = Cstruct.get_uint8 buf (offset + 2) in
        let cmd_val = Cstruct.get_uint8 buf (offset + 3) in
        read_entries buf (count - 1) ({ id_val; inter_val; cmd_val } :: lst) (offset + 4)
  in
  { option_type; length; count; entries = read_entries buf count [] 4 }

let to_cstruct t =
  let buf = Cstruct.create t.length in
  Cstruct.set_uint8 buf 0 t.option_type;
  (* Set option type *)
  Cstruct.set_uint8 buf 1 t.length;
  (* Set length *)
  Cstruct.BE.set_uint16 buf 2 t.count;
  (* Set number of entries *)
  (* Write each entry *)
  let rec write_entries buf entries offset =
    match entries with
    | [] -> buf
    | { id_val; inter_val; cmd_val } :: rest ->
        Cstruct.BE.set_uint16 buf offset id_val;
        Cstruct.set_uint8 buf (offset + 2) inter_val;
        Cstruct.set_uint8 buf (offset + 3) cmd_val;
        write_entries buf rest (offset + 4)
  in
  write_entries buf t.entries 4

type ipv4_option =
  | End_of_list (* Option type 0 *)
  | No_operation (* Option type 1 *)
  | Custom_option of Cstruct.t (* Your custom option with its own Cstruct *)
  | Unknown_option of int * Cstruct.t (* Unknown option type with raw Cstruct *)

let update_node_info ip custom_option =
  let entries = custom_option.entries in
  let rec update_node_from_entries entries =
    match entries with
    | [] -> ()
    | { id_val; inter_val; cmd_val } :: tail when cmd_val = cmd_REGISTER ->
        map_ip := add_to_map ip id_val inter_val map_ip;
        update_node_from_entries tail
    | _ :: tail -> update_node_from_entries tail
  in
  update_node_from_entries entries

let rec read_and_append_custom_opt ip_src ip_dst options_cs (node_id, node_inter_num) =
  (* (match IpV4Map.find_opt ip_dst !map_ip with
  | Some (id_val, inter_val) -> 
    log_map !map_ip;
    Logs.err (fun m -> m "ip_dst: %s - id_val: %d - inter_val: %d" (Ipaddr.V4.to_string ip_dst) id_val inter_val);
    ();
  | None -> 
    Logs.err (fun m -> m "ip_dst: %s: Not found" (Ipaddr.V4.to_string ip_dst));
    ()); *)

  match IpV4Map.find_opt ip_dst !map_ip with
  | Some (id_val, inter_val) -> (
      if Cstruct.length options_cs = 0 then (
        (* Logs.err (fun m -> m "Save my own node ID: %d and interface number %d and append DROP id_val: %d - inter_val: %d" node_id node_inter_num id_val inter_val); *)
        (* Logs.err (fun m -> m "ip_src: %s - ip_dst %s" (Ipaddr.V4.to_string ip_src) (Ipaddr.V4.to_string ip_dst)); *)
    
        map_ip := add_to_map ip_src node_id node_inter_num map_ip;
        let custom_opt = create id_val inter_val cmd_DECISION_DROP in
        to_cstruct custom_opt)
      else
        (* Get the option type (first byte) *)
        let option_type = Cstruct.get_uint8 options_cs 0 in
        match option_type with
        | 0 ->
            (* End of Option List *)
            let custom_opt = create 0 255 0 in
            let custom_opt_cs = to_cstruct custom_opt in
            Logs.debug (fun m -> m "Not found and add first custom");
            Cstruct.append custom_opt_cs options_cs
        | 1 ->
            (* No Operation (NOP) *)
            (* Move to the next option (1 byte length for NOP) *)
            Cstruct.append (Cstruct.sub options_cs 0 1) (read_and_append_custom_opt ip_src ip_dst (Cstruct.shift options_cs 1) (node_id, node_inter_num))
        | 31 ->
            (* Get custom option byte length *)
            let option_length = Cstruct.get_uint8 options_cs 1 in
            (* Read custom option bytes *)
            let custom_opt_cs = Cstruct.sub options_cs 0 option_length in
            (* Convert custom option bytes to t *)
            let custom_option = of_cstruct custom_opt_cs in
            (* Update opportunistic info from non-VC node *)
            update_node_info ip_src custom_option;
            (* Append new custom option to t *)
            let new_custom_option = add_entry custom_option id_val inter_val cmd_DECISION_DROP in
            (* Convert new custom option to bytes *)
            let new_custom_option_cs = to_cstruct new_custom_option in
            Logs.debug (fun m -> m "found and add new custom");
            (* Append the new custom option bytes to the rest of option bytes*)
            Cstruct.append new_custom_option_cs (Cstruct.shift options_cs option_length)
        | _ ->
            (* For other options, copy them unchanged and move to the next option *)
            let option_length = Cstruct.get_uint8 options_cs 1 in
            let current_option = Cstruct.sub options_cs 0 option_length in
            Cstruct.append current_option (read_and_append_custom_opt ip_src ip_dst (Cstruct.shift options_cs option_length) (node_id, node_inter_num)))
  | None ->
      (* If I was the last node can handle the attack, embed decision for my own firewall *)
      map_ip := add_to_map ip_dst node_id node_inter_num map_ip;
      Logs.err (fun m -> m "I saw an attack but did not know the source!! save my own node ID: %d and interface number" node_id);
      (* options_cs *)
      read_and_append_custom_opt ip_src ip_dst options_cs (node_id, node_inter_num)

let rec read_custom_opt ip_src options_cs (node_id, node_inter_num) =
  (if Cstruct.length options_cs = 0 then (
    Logs.debug (fun m -> m "Save my own node ID: %d and interface number %d" node_id node_inter_num);
    map_ip := add_to_map ip_src node_id node_inter_num map_ip)
  else
    (* Get the option type (first byte) *)
    let option_type = Cstruct.get_uint8 options_cs 0 in
    match option_type with
    | 0 ->
        (* End of Option List *)
        ()
    | 1 ->
        (* No Operation (NOP) *)
        read_custom_opt ip_src (Cstruct.shift options_cs 1) (node_id, node_inter_num)
    | 31 ->
        (* Get custom option byte length *)
        let option_length = Cstruct.get_uint8 options_cs 1 in
        (* Read custom option bytes *)
        let custom_opt_cs = Cstruct.sub options_cs 0 option_length in
        (* Convert custom option bytes to t *)
        let custom_option = of_cstruct custom_opt_cs in
        (* Update opportunistic info from non-VC node *)
        update_node_info ip_src custom_option;
        (* Logs.err (fun m -> m "Added new node info"); *)
        (* print_t custom_option; *)
        (* Continue to search for option type 31 (If this is the case, we have a bug :D) *)
        (* read_custom_opt ip_src (Cstruct.shift options_cs option_length) (node_id, node_inter_num) *)
    | _ ->
        (* For other options, copy them unchanged and move to the next option *)
        let option_length = Cstruct.get_uint8 options_cs 1 in
        read_custom_opt ip_src (Cstruct.shift options_cs option_length) (node_id, node_inter_num)
  );
  (* match IpV4Map.find_opt ip_src !map_ip with
  | Some (id_val, inter_val) -> 
    log_map !map_ip;
    Logs.err (fun m -> m "ip_src: %s - id_val: %d - inter_val: %d" (Ipaddr.V4.to_string ip_src) id_val inter_val);
    ();
  | None -> 
    Logs.err (fun m -> m "ip_src: %s: Not found" (Ipaddr.V4.to_string ip_src));
    (); *)
