open Lwt.Infix

module Make
      (Public_net: Mirage_net.S) (Private_net: Mirage_net.S)
      (Public_ethernet : Ethernet.S) (Private_ethernet : Ethernet.S)
  = struct

  (* configure logs, so we can use them later *)
  let log = Logs.Src.create "listener" ~doc:"Listener for public and private interfaces"
  module Log = (val Logs.src_log log : Logs.LOG)

  (* output callbacks we need *)
  let output_public public_netif packet =
    let len = Cstruct.length packet in
    Public_net.write public_netif ~size:len (fun b -> Cstruct.blit packet 0 b 0 len ; len) >|= function
    | Ok () -> ()
    | Error e ->
      Log.warn (fun f -> f "netif write errored %a" Public_net.pp_error e)

  (* Forward the (dest, packet) [packet] to the private interface, using [dest] to understand how to route *)
  let output_private private_netif packet =
    let len = Cstruct.length packet in
    Private_net.write private_netif ~size:len (fun b -> Cstruct.blit packet 0 b 0 len ; len) >|= function
    | Ok () -> ()
    | Error e ->
      Log.warn (fun f -> f "netif write errored %a" Private_net.pp_error e)

  let output_ether_private private_ethernet mac_src mac_dst packet =
    Private_ethernet.write private_ethernet ~src:mac_src mac_dst `IPv4 (fun b ->
      let len = Cstruct.length packet in
      Cstruct.blit packet 0 b 0 len ;len) >>= function
      | Error e ->
        Log.err (fun f -> f "Failed to send packet from private interface: %a" Private_ethernet.pp_error e);
        Lwt.return_unit
      | Ok () -> Lwt.return_unit

  (* we need to establish listeners for the and public interfaces *)
  (* we're interested in all traffic to the physical interface; we'd like to
     send ARP traffic to the normal ARP listener and responder,
     handle ipv4 traffic with the functions we've defined above for detection,
     and ignore all ipv6 traffic. *)
  let listen_public public_netif private_netif private_ethernet inter_num =
    let header_size = Ethernet.Packet.sizeof_ethernet
    and input frame = (* Takes an ethernet packet and send it to the relevant callback *)
        match Ethernet.Packet.of_cstruct frame with
        | Ok (header, payload) ->
          begin
            match header.Ethernet.Packet.ethertype with
            | `ARP -> output_private private_netif frame
            | `IPv4 -> Ids.detect_and_output_private payload header frame inter_num (output_private private_netif) (output_ether_private private_ethernet)
            | _ -> Lwt.return_unit
          end
        | Error s ->
          Log.debug (fun f -> f "Forward Ethernet frame if I cannot parse %s" s);
          output_private private_netif frame;
    in
    Public_net.listen ~header_size public_netif input >>= function
    | Error e -> Log.debug (fun f -> f "public interface stopped: %a" Public_net.pp_error e); Lwt.return_unit
    | Ok () -> Log.debug (fun f -> f "public interface terminated normally");
      Lwt.return_unit

  let listen_private public_netif private_netif =
    let header_size = Ethernet.Packet.sizeof_ethernet
    and input frame =
      output_public public_netif frame
    in
    Private_net.listen ~header_size private_netif input >>= function
    | Error e -> Log.debug (fun f -> f "private interface stopped: %a" Private_net.pp_error e); Lwt.return_unit
    | Ok () -> Log.debug (fun f -> f "private interface terminated normally"); Lwt.return_unit

end
