let log = Logs.Src.create "rules" ~doc:"IDS rules management"

module Log = (val Logs.src_log log : Logs.LOG)

(* flow option *)
type flow = TO_CLIENT | TO_SERVER | FROM_CLIENT | FROM_SERVER | ESTABLISHED | NOT_ESTABLISHED | STATELESS | ONLY_STREAM | ONLY_FRAG | NO_FRAG

(* content option *)
let content_of_string content_string =
  let is_hex = String.starts_with ~prefix:"|" content_string in
  let parts = String.split_on_char '|' content_string in
  let rec process_hex hex_parts acc =
    match hex_parts with
    | [] -> acc
    | "" :: rest -> process_hex rest acc (* Skip empty segments *)
    | part :: rest ->
        let segment = Cstruct.of_hex part in
        process_hex rest (Cstruct.append acc segment)
  in
  let rec process_parts is_hex parts acc =
    match parts with
    | [] -> acc
    | "" :: rest -> process_parts true rest acc (* Skip empty segments *)
    | part :: rest ->
        if is_hex then
          let segment =
            let hex_char = String.split_on_char ' ' part in
            process_hex hex_char Cstruct.empty
          in
          process_parts false rest (Cstruct.append acc segment)
        else
          let segment = Cstruct.of_string part in
          process_parts true rest (Cstruct.append acc segment)
  in
  process_parts is_hex parts Cstruct.empty

(* ByeTest option: <num of bytes> | <variable_name>, [!]<operator>, <test value>, <offset> [,relative]*)
(* [,<endian>][, string, <num type>][, dce][, bitmask <bitmask value>] *)
type operator = L | G | EQ | LE | GE | AND | OR | NL | NG | NEQ | NLE | NGE | NAND | NOR
type endian = BIG | LITTLE
type num_type = HEX | DEC | OCT

type byte_test = {
  num_of_bytes : int;
  variable_name : string option;
  operator : operator;
  test_value : Cstruct.t;
  offset : int;
  relative : bool;
  endian : endian option;
  string_num_type : (string * num_type) option;
  dce : bool;
  bitmask : Cstruct.t option;
}

type offset = Offset of int
type depth = Depth of int

(* Flowbits option *)
type flowbits_opt = SET | IS_SET | UN_SET | IS_NOT_SET | NO_ALERT

(* Reference meta keyword *)
type ref_type = CURL | CVE

(* Classtype meta keyword *)
type classtype = ATTEMPTED_ADMIN

type rule_opt =
  | Flow of flow list
  | Content of (Cstruct.t * depth * offset)
  | ByteTest of byte_test
  | Pcre_pattern of Re.re
  | Flowbits of (flowbits_opt * string)

type meta_kwd = Msg of string | Reference of (ref_type * string) | Classtype of classtype | Sid of int | Rev of int
type decision = NORMAL | ALERT
type protocol = [ `SMB | `TCP | `UDP | `ICMP ]

type rule = {
  src : Ipaddr.V4.Prefix.t;
  psrc : int;
  dst : Ipaddr.V4.Prefix.t;
  pdst : int;
  proto : protocol option;
  option : rule_opt list option;
  meta : meta_kwd list option;
  action : decision option;
}

(* A match is either the target is joker or the ip is in the target network *)
let match_ip ip target = target = Ipaddr.V4.Prefix.global || Ipaddr.V4.Prefix.mem ip target

type t = { mutable l : rule list; default : bool }

(* NOTE: No verification on the grammar of rule option *)
let init default =
  {
    l =
      [
        {
          src = Ipaddr.V4.Prefix.global;
          psrc = 0;
          dst = Ipaddr.V4.Prefix.global;
          pdst = 0;
          proto = Some `SMB;
          option =
            Some
              [
                Flow [ TO_SERVER; ESTABLISHED ];
                Content (content_of_string "|FF|SMB2|00 00 00 00|", Depth 9, Offset 4);
                ByteTest
                  {
                    num_of_bytes = 2;
                    variable_name = None;
                    operator = G;
                    test_value = Cstruct.of_hex "0008";
                    offset = 52;
                    relative = true;
                    endian = Some LITTLE;
                    string_num_type = None;
                    dce = false;
                    bitmask = None;
                  };
                Pcre_pattern (Re.Pcre.regexp ~flags:[ `DOTALL ] "\\xFFSMB2\\x00\\x00\\x00\\x00.{52}(?:\\x04|\\x09|\\x0A|\\x0B|\\x0C|\\x0E|\\x11)\\x00");
                Flowbits (SET, "SMB.Trans2.SubCommand.Unimplemented");
              ];
          meta =
            Some
              [
                Msg "ATTACK [PTsecurity] Unimplemented Trans2 Sub-Command code. Possible ETERNALBLUE (WannaCry, Petya) tool";
                Reference (CURL, "msdn.microsoft.com/en-us/library/ee441654.aspx");
                Classtype ATTEMPTED_ADMIN;
                Reference (CURL, "github.com/ptresearch/AttackDetection");
                Sid 10001254;
                Rev 2;
              ];
          action = Some ALERT;
        };
        {
          src = Ipaddr.V4.Prefix.global;
          psrc = 0;
          dst = Ipaddr.V4.Prefix.global;
          pdst = 12349;
          proto = Some `TCP;
          option = None;
          meta = None;
          action = Some ALERT;
        };
      ];
    default;
  }

type record = { is_establish : bool; is_client : bool }

(* TODO: The record should be updated with expired time
   NOTE: The record won't work if the request and response packets are in different flows *)
type connection_state = COMPLETE_HANDSHAKE_CLIENT | COMPLETE_HANDSHAKE_SERVER | INITIATE_SENT | SYN_ACK_SENT | CLOSED

module RecordTuple = struct
  type t = { record_src : Ipaddr.V4.t; record_psrc : int; record_dst : Ipaddr.V4.t; record_pdst : int; record_proto : int }

  let pp f { record_src; record_psrc; record_dst; record_pdst; record_proto } =
    Fmt.pf f "{record_src=%s;record_psrc=%d;record_dst=%s;record_pdst=%d;record_proto=%d}" (Ipaddr.V4.to_string record_src) record_psrc
      (Ipaddr.V4.to_string record_dst) record_pdst record_proto

  let compare = compare
end

module RecordMap = Map.Make (RecordTuple)

let map_record = ref RecordMap.empty

let record_flow_tcp : Ipv4_packet.t -> int -> int -> bool -> bool -> bool -> bool -> unit =
 fun ipv4_hdr psrc pdst syn ack rst fin ->
  let key : RecordTuple.t =
    { record_src = ipv4_hdr.src; record_psrc = psrc; record_dst = ipv4_hdr.dst; record_pdst = pdst; record_proto = ipv4_hdr.Ipv4_packet.proto }
  in
  let key2 : RecordTuple.t =
    { record_src = ipv4_hdr.dst; record_psrc = pdst; record_dst = ipv4_hdr.src; record_pdst = psrc; record_proto = ipv4_hdr.Ipv4_packet.proto }
  in
  match (syn, ack, rst, fin) with
  | true, false, false, false ->
      let flowbit_name = match RecordMap.find_opt key !map_record with None -> "" | Some (_, value) -> value in
      map_record := RecordMap.add key (INITIATE_SENT, flowbit_name) !map_record;
      Log.debug (fun f -> f "New record %a %s with Flowbits: %s" RecordTuple.pp key "SYN_SENT" flowbit_name)
  | true, true, false, false -> (
      match RecordMap.find_opt key2 !map_record with
      | Some (INITIATE_SENT, flowbit_name) ->
          map_record := RecordMap.add key (SYN_ACK_SENT, flowbit_name) !map_record;
          Log.debug (fun f -> f "New record %a %s with Flowbits: %s" RecordTuple.pp key "SYN_ACK_SENT" flowbit_name)
      | _ -> ())
  | false, true, false, false -> (
      match RecordMap.find_opt key2 !map_record with
      | Some (SYN_ACK_SENT, flowbit_name) ->
          map_record := RecordMap.add key (COMPLETE_HANDSHAKE_CLIENT, flowbit_name) !map_record;
          Log.debug (fun f -> f "New record %a %s with Flowbits: %s" RecordTuple.pp key "COMPLETE_HANDSHAKE_CLIENT" flowbit_name);
          map_record := RecordMap.add key2 (COMPLETE_HANDSHAKE_SERVER, flowbit_name) !map_record;
          Log.debug (fun f -> f "New record %a %s with Flowbits: %s" RecordTuple.pp key2 "COMPLETE_HANDSHAKE_SERVER" flowbit_name)
      | _ -> ())
  | false, true, true, false | false, true, false, true -> (
      match RecordMap.find_opt key !map_record with
      | Some (_, "") ->
          map_record := RecordMap.remove key !map_record;
          map_record := RecordMap.remove key2 !map_record
      | Some (_, value) ->
          map_record := RecordMap.add key (CLOSED, value) !map_record;
          map_record := RecordMap.add key2 (CLOSED, value) !map_record;
          Log.debug (fun f -> f "New record %a %s" RecordTuple.pp key "CLOSED")
      | None -> ())
  | _ -> ()

(* Record UDP currently dont care about broadcast packets *)
let record_flow_udp : Ipv4_packet.t -> int -> int -> unit =
 fun ipv4_hdr psrc pdst ->
  let key : RecordTuple.t =
    { record_src = ipv4_hdr.src; record_psrc = psrc; record_dst = ipv4_hdr.dst; record_pdst = pdst; record_proto = ipv4_hdr.Ipv4_packet.proto }
  in
  let key2 : RecordTuple.t =
    { record_src = ipv4_hdr.dst; record_psrc = pdst; record_dst = ipv4_hdr.src; record_pdst = psrc; record_proto = ipv4_hdr.Ipv4_packet.proto }
  in
  match RecordMap.find_opt key !map_record with
  | None -> (
      match RecordMap.find_opt key2 !map_record with
      | None ->
          let flowbit_name = match RecordMap.find_opt key !map_record with None -> "" | Some (_, value) -> value in
          map_record := RecordMap.add key (INITIATE_SENT, flowbit_name) !map_record;
          Log.debug (fun f -> f "New record %a %s with Flowbits: %s" RecordTuple.pp key "INITIATE_SENT" flowbit_name)
      | Some (INITIATE_SENT, flowbit_name) ->
          map_record := RecordMap.add key2 (COMPLETE_HANDSHAKE_CLIENT, flowbit_name) !map_record;
          Log.debug (fun f -> f "New record %a %s" RecordTuple.pp key2 "COMPLETE_HANDSHAKE_CLIENT");
          map_record := RecordMap.add key (COMPLETE_HANDSHAKE_SERVER, flowbit_name) !map_record;
          Log.debug (fun f -> f "New record %a %s with Flowbits: %s" RecordTuple.pp key "COMPLETE_HANDSHAKE_SERVER" flowbit_name)
      | _ -> ())
  | _ -> ()

let is_match_option p_src p_psrc p_dst p_pdst p_proto r_opt packet =
  let key : RecordTuple.t = { record_src = p_src; record_psrc = p_psrc; record_dst = p_dst; record_pdst = p_pdst; record_proto = p_proto } in
  match r_opt with
  | Some options when List.length options > 0 ->
      (* Check specified options in turn *)
      let current_offset = 0 in
      let rec is_match_opts opts off =
        match opts with
        | opt :: rest ->
            let rs, relative_off =
              match opt with
              | Flow flows ->
                  (* NOTE: Currently only care about established and to_server *)
                  let rec is_match_flows fls =
                    match fls with
                    | ESTABLISHED :: rest -> (
                        match RecordMap.find_opt key !map_record with
                        | Some (COMPLETE_HANDSHAKE_CLIENT, _) | Some (COMPLETE_HANDSHAKE_SERVER, _) ->
                            Log.debug (fun f -> f "ESTABLISHED matched");
                            is_match_flows rest
                        | _ -> (false, off))
                    | TO_SERVER :: rest -> (
                        (* TODO: Not sure how suricata decide whenever a flow is to_server or to_client -> How to know? *)
                        match RecordMap.find_opt key !map_record with
                        | Some (INITIATE_SENT, _) | Some (COMPLETE_HANDSHAKE_CLIENT, _) ->
                            Log.debug (fun f -> f "TO_SERVER matched");
                            is_match_flows rest
                        | _ -> (false, off))
                    | [] -> (true, off)
                    | _ -> (false, off)
                  in
                  is_match_flows flows
              | Content (content_to_match, Depth depth, Offset offset) ->
                  let temp_off = off + offset in
                  let content_packet = Cstruct.sub packet temp_off depth in
                  Log.debug (fun f -> f "Content matched: %b" (Cstruct.equal content_to_match content_packet));
                  (Cstruct.equal content_to_match content_packet, temp_off + depth)
              | ByteTest { num_of_bytes; variable_name; operator; test_value; offset; relative; endian; string_num_type = _; dce = _; bitmask = _ } -> (
                  (* NOTE: Currently, we dont care about variable_name, string_num_typ, dce, bitmask due to current testbed *)
                  match variable_name with
                  | None -> (
                      let temp_off = match relative with false -> 0 + offset | true -> off + offset in
                      let content_packet = Cstruct.sub packet temp_off num_of_bytes in
                      let content_packet = match endian with Some LITTLE -> Cstruct.rev content_packet | _ -> content_packet in
                      match operator with
                      (* TO DO: bitwise AND, OR, NAND, NOR still missing *)
                      | AND | OR | NAND | NOR -> (false, temp_off + num_of_bytes)
                      | _ -> (
                          (* For other operators, perform the comparison once and reuse the result *)
                          let compare_result = Cstruct.compare content_packet test_value in
                          match (operator, compare_result) with
                          | (L | LE | NG | NGE | NEQ), -1 | (G | GE | NL | NLE | NEQ), 1 | (EQ | LE | GE | NL | NG), 0 ->
                              Log.debug (fun f ->
                                  f "Byte_test matched: %s %d %s" (Cstruct.to_hex_string content_packet) compare_result (Cstruct.to_hex_string test_value));
                              (true, temp_off + num_of_bytes)
                          | _ -> (false, temp_off + num_of_bytes)))
                  | _ -> (false, off))
              | Pcre_pattern pcre ->
                  let packet_str = Cstruct.to_string packet in
                  let rs = Re.Pcre.pmatch ~rex:pcre packet_str in
                  Log.debug (fun f -> f "Pcre matched: %b" rs);
                  (rs, off)
              | Flowbits flowbit -> (
                  (* TO DO: Not sure if multiple ports between two endponits are considered as the same flow to set the same state? *)
                  match (flowbit, RecordMap.find_opt key !map_record) with
                  | (SET, name), Some (connection_state, _) ->
                      map_record := RecordMap.add key (connection_state, name) !map_record;
                      Log.debug (fun f -> f "Update record %a with Flowbits: %s" RecordTuple.pp key name);
                      (true, off)
                  | (UN_SET, name), Some (connection_state, flowbit_name) when flowbit_name = name ->
                      map_record := RecordMap.add key (connection_state, "") !map_record;
                      Log.debug (fun f -> f "Update record %a with Flowbits: UN_SET" RecordTuple.pp key);
                      (true, off)
                  | (IS_SET, name), Some (_, flowbit_name) when flowbit_name = name ->
                      Log.debug (fun f -> f "Flowbits matched");
                      (true, off)
                  | (IS_SET, _), _ -> (false, off)
                  | (IS_NOT_SET, name), Some (_, flowbit_name) when flowbit_name != name ->
                      Log.debug (fun f -> f "Flowbits matched");
                      (true, off)
                  | (IS_NOT_SET, _), _ -> (false, off)
                  | _ -> (true, off))
              (* | Flowbits of (string * string) *)
            in
            rs && is_match_opts rest relative_off
        | [] -> true
      in
      is_match_opts options current_offset
  | _ -> true

let sign_smb_cs = Cstruct.of_string "\xFFSMB"
let sign_smb2_cs = Cstruct.of_string "\xFESMB"

let is_match_smb packet =
  (* NetBios session layer has 4 bytes, then 4 bytes SMB signature in SMB header *)
  if Cstruct.length packet > 7 then
    let p_smb_cs = Cstruct.sub packet 4 4 in
    (* If the packet is truely an SMB one, check opt if it is specified *)
    if Cstruct.equal p_smb_cs sign_smb_cs || Cstruct.equal p_smb_cs sign_smb2_cs then (
      Log.debug (fun f -> f "Found an SMB packet");
      true)
    else false
  else false

let is_match_proto p_proto proto =
  match (p_proto, proto) with Some `TCP, Some `TCP -> true | Some `UDP, Some `UDP -> true | Some `ICMP, Some `ICMP -> true | _ -> false

(* Takes an ipv4 header [ipv4_hdr] and the whole IPv4 [packet] (containing the header).
   We want to alert any packet matching the [filters] list
   NOTE: Currently, we only care about SMB rules with option, therefore no rule option check for other rules *)
let is_match_rule : t -> Ipv4_packet.t * Cstruct.t -> bool =
 fun t (ipv4_hdr, packet) ->
  let rec apply_rules : bool -> Ipaddr.V4.t -> Ipaddr.V4.t -> int -> int -> int -> Cstruct.t -> rule list -> bool =
   fun default p_src p_dst p_proto p_psrc p_pdst packet ids_rules ->
    match ids_rules with
    (* If the list is empty -> apply default (true or false) action *)
    | [] -> default
    (* If the packet matches (SMB), conditions, consider rule option and has an alert action *)
    | { src; psrc; dst; pdst; proto = Some `SMB; option; meta = _; action = Some ALERT } :: _
      when match_ip p_src src && match_ip p_dst dst
           && (not (is_match_proto (Ipv4_packet.Unmarshal.int_to_protocol p_proto) (Some `ICMP)))
           && (psrc = 0 || p_psrc = psrc)
           && (pdst = 0 || p_pdst = pdst)
           && is_match_smb packet
           && is_match_option ipv4_hdr.src p_psrc ipv4_hdr.dst p_pdst ipv4_hdr.Ipv4_packet.proto option packet ->
        Log.warn (fun f -> f "Alert a packet from %a to %a..." Ipaddr.V4.pp p_src Ipaddr.V4.pp p_dst);
        true
    (* If the packet matches (ICMP, TCP, UDP), conditions, no consider rule option and has an alert action *)
    | { src; psrc; dst; pdst; proto; option = _; meta = _; action = Some ALERT } :: _
      when match_ip p_src src && match_ip p_dst dst
           && (proto = None || is_match_proto (Ipv4_packet.Unmarshal.int_to_protocol p_proto) proto)
           && (psrc = 0 || p_psrc = psrc)
           && (pdst = 0 || p_pdst = pdst) ->
        Log.warn (fun f -> f "Alert a packet from %a to %a..." Ipaddr.V4.pp p_src Ipaddr.V4.pp p_dst);
        true
    (* Or finally the packet does not match the condition *)
    | _ :: tail -> apply_rules default p_src p_dst p_proto p_psrc p_pdst packet tail
  in

  match ipv4_hdr.Ipv4_packet.proto with
  | 6 -> (
      (* TCP *)
      match Tcp.Tcp_packet.Unmarshal.of_cstruct packet with
      | Result.Error s ->
          Logs.err (fun m -> m "Can't parse TCP packet: %s" s);
          false
      | Result.Ok (tcp_hdr, payload) ->
          record_flow_tcp ipv4_hdr tcp_hdr.src_port tcp_hdr.dst_port tcp_hdr.syn tcp_hdr.ack tcp_hdr.rst tcp_hdr.fin;
          apply_rules t.default ipv4_hdr.src ipv4_hdr.dst ipv4_hdr.Ipv4_packet.proto tcp_hdr.src_port tcp_hdr.dst_port payload t.l)
  | 17 -> (
      (* UDP *)
      match Udp_packet.Unmarshal.of_cstruct packet with
      | Result.Error s ->
          Logs.err (fun m -> m "Can't parse UDP packet: %s" s);
          false
      | Result.Ok (udp_hdr, payload) ->
          record_flow_udp ipv4_hdr udp_hdr.src_port udp_hdr.dst_port;
          apply_rules t.default ipv4_hdr.src ipv4_hdr.dst ipv4_hdr.Ipv4_packet.proto udp_hdr.src_port udp_hdr.dst_port payload t.l)
  | 1 -> false (* ICMP *)
  | _ -> false
