import netifaces
import time
import os
import queue
import signal
import subprocess
import threading

debug = False
NAME_INF = "vmnet" if debug else "ens"
MIN_INTER_NUMBER = 1 if debug else 3
MAX_INTER_NUMBER = 2 if debug else 13
PHYSICAL_INTERFACE = {}
for index in range(MIN_INTER_NUMBER, MAX_INTER_NUMBER):
    PHYSICAL_INTERFACE[index] = NAME_INF + str(index + 1)

map_bridge_outside = {}
map_inf_inside = {}
map_inf_ip = {}
interface_tap_id = 4
my_id = 0

DIR = "/home/node/Desktop/gns3_unikernel_testbed"


def create_tuntap_interface(interface_name):
    """
    Create a tuntap mode tap interface and activate it.
    :param interface_name: Name of the tap interface
    """
    subprocess.run(["sudo", "ip", "tuntap", "add", interface_name, "mode", "tap"], check=True)
    subprocess.run(["sudo", "ip", "link", "set", "dev", interface_name, "up"], check=True)


def create_bridge(bridge_name):
    """
    Create a bridge, turn it up, and enable STP using brctl.
    :param bridge_name: Name of the bridge
    """
    subprocess.run(["sudo", "brctl", "addbr", bridge_name], check=True)
    subprocess.run(["sudo", "ip", "link", "set", bridge_name, "up"], check=True)
    subprocess.run(["sudo", "brctl", "stp", bridge_name, "on"], check=True)


def add_interface_to_bridge(bridge_name, interface_name):
    """
    Add an interface to a bridge.
    :param bridge_name: Name of the bridge
    :param interface_name: Name of the interface
    """
    subprocess.run(["sudo", "brctl", "addif", bridge_name, interface_name], check=True)


def is_interface_down(interface):
    """
    Check if the network interface is down.
    :param interface: Name of the network interface to check
    :return: True if the interface is down, False otherwise
    """
    try:
        result = subprocess.run(
            ["ip", "link", "show", interface],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True
        )
        return "state DOWN" in result.stdout
    except subprocess.CalledProcessError:
        print(f"Failed to check status of interface {interface}.")
        return False


def delete_tuntap_interface(interface_name):
    """
    Delete a tuntap mode tap interface.
    :param interface_name: Name of the tap interface to delete
    """
    # Bring the interface down before deleting it
    subprocess.run(["sudo", "ip", "link", "set", "dev", interface_name, "down"], check=True)

    subprocess.run(["sudo", "ip", "tuntap", "del", interface_name, "mode", "tap"], check=True)


def delete_bridge(bridge_name):
    """
    Delete a bridge.
    :param bridge_name: Name of the bridge
    """
    subprocess.run(["sudo", "ip", "link", "set", bridge_name, "down"], check=True)
    subprocess.run(["sudo", "brctl", "delbr", bridge_name], check=True)


def delete_interface_from_bridge(bridge_name, interface_name):
    """
    Delete an interface from a bridge.
    :param bridge_name: Name of the bridge
    :param interface_name: Name of the interface
    """
    # Note: 'brctl delif' does not require the bridge to be down
    subprocess.run(["sudo", "brctl", "delif", bridge_name, interface_name], check=True)


def forward_to_queue_rule(inter_name, queue_num):
    """
    Run iptables command to forward traffic to NFQUEUE.
    :param inter_name: The name of the interface, e.g., "ens4"
    :param queue_num: The NFQUEUE number to be used
    """
    command = ["sudo", "iptables", "-A", "FORWARD", "-i", inter_name, "-j", "NFQUEUE", "--queue-num", str(queue_num)]
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        print(f"Error executing command {command}: {e}")


def delete_forward_to_queue_rule(inter_name, queue_num):
    """
    Run iptables command to delete forward rule to NFQUEUE.
    :param inter_name: The name of the interface, e.g., "ens4"
    :param queue_num: The NFQUEUE number to be used
    """
    command = ["sudo", "iptables", "-D", "FORWARD", "-i", inter_name, "-j", "NFQUEUE", "--queue-num", str(queue_num)]
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        cmd_err = " ".join(command)
        print(f"Error executing command {cmd_err}: {e}")


def drop_layer2_VC_packet_rule(mac_address):
    """
    Run ebtables command to drop bridge packets sent by vc algorithms (avoid forwarding neighbors' VC messages).
    :param mac_address: the output address to be dropped
    """
    command = ["sudo", "ebtables", "-I", "FORWARD", "-d", mac_address, "-j", "DROP"]
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        cmd_err = " ".join(command)
        print(f"Error executing command {cmd_err}: {e}")


def drop_internet_and_allow_VC_server_rule(inf_to_internet, server_port):
    """
    Run iptables command to drop packets target outside internet.
    :param server_port: port of the VC server running on the main host
    :param inf_to_internet: the interface that connect to the main host (internet)
    """

    commands = [["sudo", "iptables", "-t", "nat", "-A", "POSTROUTING", "-o", inf_to_internet, "-j", "MASQUERADE"],
                ["sudo", "iptables", "-A", "FORWARD", "-o", inf_to_internet, "-p", "tcp", "--dport", str(12348), "-j", "ACCEPT"],
                ["sudo", "iptables", "-A", "FORWARD", "-o", inf_to_internet, "-j", "DROP"],
                ["sudo", "iptables", "-A", "OUTPUT", "-o", inf_to_internet, "-p", "tcp", "--dport", str(server_port), "-j", "ACCEPT"],
                ["sudo", "iptables", "-A", "OUTPUT", "-o", inf_to_internet, "-j", "DROP"]]

    for command in commands:
        try:
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as e:
            cmd_err = " ".join(command)
            print(f"Error executing command {cmd_err}: {e}")


def mark_mangle_forward_rule(inter_name, mark_value):
    """
    Add a rule to the mangle table to mark packets from a specific interface.
    :param bridge_name: The name of the bridge (e.g., "br0")
    :param inter_name: The name of the physical device interface (e.g., "eth0")
    :param mark_value: The mark value to set (e.g., 1)
    """
    command = ["sudo", "iptables", "-t", "mangle", "-A", "FORWARD", "-i", inter_name, "-j", "MARK", "--set-mark", str(mark_value)]
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        cmd_err = " ".join(command)
        print(f"Error executing command {cmd_err}: {e}")


def delete_mangle_mark_rule(inter_name, mark_value):
    """
    Delete a rule from the mangle table that marks packets from a specific interface.
    :param inter_name: The name of the physical device interface (e.g., "eth0")
    :param mark_value: The mark value to remove (e.g., 1)
    """
    command = ["sudo", "iptables", "-t", "mangle", "-D", "FORWARD", "-i", inter_name, "-j", "MARK", "--set-mark", str(mark_value)]
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        cmd_err = " ".join(command)
        print(f"Error executing command {cmd_err}: {e}")


def enable_br_netfilter():
    """
    Load the br_netfilter kernel module using modprobe.
    """
    try:
        subprocess.run(["sudo", "modprobe", "br_netfilter"], check=True)
        # print("Successfully loaded br_netfilter module.")
    except subprocess.CalledProcessError as e:
        print(f"Error loading br_netfilter module: {e}")


class EventDrivenQueue:
    def __init__(self, process_function):
        """
        Initialize the EventDrivenQueue.
        :param process_function: The function to call whenever new data is added to the queue.
        """
        self.queue = queue.Queue()
        self.process_function = process_function
        self.worker_thread = threading.Thread(target=self._worker, daemon=True)
        self.worker_thread.start()

    def _worker(self):
        """Worker thread that processes data as it is added to the queue."""
        while True:
            data = self.queue.get()  # Wait for data
            try:
                self.process_function(data)  # Process the data
            finally:
                self.queue.task_done()

    def put(self, data):
        """
        Add data to the queue for processing.
        :param data: The data to process.
        """
        self.queue.put(data)


def run_firewall_solo5_hvt(private, public):
    """
    Run the solo5-hvt command with the given parameters.
    :param private: The private network interface
    :param public: The public network interface
    :return: The output of the command or an error message
    """

    solo5_hvt_path = "/home/node/.opam/4.14.2/bin/solo5-hvt"
    hvt_file = f"{DIR}/simple-fw/dist/simple-fw.hvt"
    command = [
        solo5_hvt_path,
        "--mem=32",
        f"--net:private={private}",
        f"--net:public={public}",
        hvt_file,
    ]

    try:
        command_str = " ".join(command)
        # print(f"{command_str}")
        subprocess.Popen(
            ['xterm', '-hold', '-e', command_str],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            stdin=subprocess.DEVNULL,
            start_new_session=True
        )
    except subprocess.CalledProcessError as e:
        cmd_err = " ".join(command)
        print(f"Error executing command {cmd_err}: {e}")


def run_ids_solo5_hvt(map_tap_inf_for_ids):
    """
    Run the solo5-hvt command with the given parameters.

    :param map_tap_inf_for_ids: map containing public and private tap intefaces for IDS
    :return: None
    """
    # Initialize the base command
    command = [
        "/home/node/.opam/4.14.2/bin/solo5-hvt",
        "--mem=32"
    ]

    # Loop through interface numbers from 4 to 13
    for i in map_tap_inf_for_ids.keys():
        inter_name, new_bridge, public_if = map_tap_inf_for_ids[i]
        private_if = map_inf_inside[i]
        # Add the --net option dynamically for each interface
        command.append(f"--net:public{i + 1}={public_if}")
        command.append(f"--net:private{i + 1}={private_if}")

    # Append the final path to the HVT file
    command.append(f"{DIR}/simple-IDS/dist/simple-ids.hvt")

    # Append node ID
    command.append(f"--node-id={my_id}")

    command_str = " ".join(command)
    # process = subprocess.Popen(['gnome-terminal', '--', 'bash', '-c', command_str])
    process = subprocess.Popen(
        ['xterm', '-hold', '-e', command_str],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        stdin=subprocess.DEVNULL,
        start_new_session=True
    )

    return process


def stop_ids_solo5_hvt(process):
    subproc_pid = None
    with os.popen(f"pgrep -f {DIR}/simple-IDS/dist/simple-ids.hvt") as proc_list:
        for pid in proc_list.readlines():
            subproc_pid = int(pid.strip())
            if subproc_pid != process.pid:
                break

    # Kill the subprocess by its PID
    if subproc_pid:
        os.kill(subproc_pid, signal.SIGTERM)
        # print(f"----------------- Stopped the IDS with PID {subproc_pid}. -----------------")
    # else:
    # print("----------------- No child process found to kill. -----------------")


def setup_dhcp(infs, ips, subnets):
    dhcp_config = "default-lease-time 7200;\n" + \
                  "max-lease-time 7200;\n"

    for i in range(len(subnets)):
        subnet = subnets[i]
        subnet_octets = subnet.split(".")
        subnet_octets[-1] = "1"
        lower_bound = ".".join(subnet_octets)
        subnet_octets[-1] = "100"
        upper_bound = ".".join(subnet_octets)
        ip = ips[i]
        dhcp_config += f"subnet {subnet} netmask 255.255.255.0 {{\n" + \
                       f"range {lower_bound} {upper_bound};\n" + \
                       f"option routers {ip};\n}}\n\n"

    print("[*] Configuring DHCP server...")
    with open("/tmp/dhcpd.conf", "w") as f:
        f.write(dhcp_config)
    subprocess.run(["sudo", "mv", "/tmp/dhcpd.conf", "/etc/dhcp/dhcpd.conf"], check=True)

    print("[*] Setting DHCP interface...")
    inf_config = f'INTERFACESv4="{" ".join(infs)}"\n'
    with open("/tmp/isc-dhcp-server", "w") as f:
        f.write(inf_config)
        f.write(f'INTERFACESv6=""\n')
    subprocess.run(["sudo", "mv", "/tmp/isc-dhcp-server", "/etc/default/isc-dhcp-server"], check=True)

    try:
        subprocess.run(["sudo", "systemctl", "enable", "isc-dhcp-server"], check=True, capture_output=True, text=True)
        subprocess.run(["sudo", "systemctl", "restart", "isc-dhcp-server"], check=True, capture_output=True, text=True)
        print("[✔] DHCP server setup completed!")
    except subprocess.CalledProcessError as e:
        print("Failed to restart DHCP server!")
        print(f"Exit Code: {e.returncode}")
        print(f"Error Output: {e.stderr}")


def disable_NetworkManager_auto_ip(inter_name):
    command = ["sudo", "nmcli", "device", "set", inter_name, "managed", "no"]
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        cmd_err = " ".join(command)
        print(f"Error executing command {cmd_err}: {e}")


def add_ip_address(inter_name, ip_address, mask):
    command = ["sudo", "ip", "addr", "add", ip_address + mask, "dev", inter_name]
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        cmd_err = " ".join(command)
        print(f"Error executing command {cmd_err}: {e}")


def delete_ip_address(inter_name, ip_address):
    command = ["sudo", "ip", "addr", "del", ip_address + "/24", "dev", inter_name]
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        cmd_err = " ".join(command)
        print(f"Error executing command {cmd_err}: {e}")


def wait_for_ip(interface_name, prefix_ip):
    while True:
        try:
            # Get the addresses for the specific interface
            addresses = netifaces.ifaddresses(interface_name)
            # Fetch the IPv4 address (AF_INET)
            ip_address = addresses[netifaces.AF_INET][0]['addr']
            if ip_address.startswith(prefix_ip):
                # print(f"IP Address assigned: {ip_address}")
                return ip_address
            else:
                raise Exception("Waiting for valid IP address")
            break
        except:
            # print("No valid IP address assigned yet. Retrying...")
            time.sleep(5)  # Wait 5 seconds before checking again


def apply_nat_rule(private_ip, public_ip):
    command = ["iptables", "-t", "nat", "-A", "PREROUTING", "-d", public_ip, "-j", "DNAT", "--to-destination", private_ip]
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        cmd_err = " ".join(command)
        print(f"Error executing command {cmd_err}: {e}")

    command = ["iptables", "-t", "nat", "-A", "POSTROUTING", "-s", private_ip, "-j", "SNAT", "--to-source", public_ip]
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        cmd_err = " ".join(command)
        print(f"Error executing command {cmd_err}: {e}")


def process_vtysh(commands, update_str):
    process = subprocess.Popen(
        ["sudo", "vtysh"],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    output, errors = process.communicate(input=commands)
    if process.returncode == 0:
        print(f"Update {update_str} successfully.")
    else:
        print(f"Failed to configure {update_str}.")


def update_ospf_rule(input_map, total_router):
    try:
        MY_BGP_AS_NUM = 65000
        subnets = []
        subnets_user = []
        ip_infs_user = []
        infs_user = []

        commands = "configure terminal"
        commands += "\nrouter ospf"
        commands += f"\npassive-interface ens3"

        bgp_commands = f"\nrouter bgp {MY_BGP_AS_NUM}"

        lo_ip = f"1.1.{my_id}.1"
        bgp_commands += f"\nbgp router-id {lo_ip}"
        add_ip_address("lo", lo_ip, "/32")

        for iface_number_str, ip in input_map.items():
            iface_name = f'ens{int(iface_number_str) + 3}'  # setup dhcp on ens inf, no need to set up on tap infs
            octets = ip.split(".")
            octets[-1] = "0"
            subnet = ".".join(octets)
            subnets.append(subnet)

            # Check if the ip address belongs to users to run dhcp server
            if subnet.startswith('123'):
                subnets_user.append(subnet)
                ip_infs_user.append(ip)
                infs_user.append(iface_name)
                commands += f"\npassive-interface {iface_name}"

                temp_octets = ip.split(".")
                temp_octets[-1] = "1"
                peer_ip = ".".join(temp_octets)
                bgp_commands += f"\nneighbor {peer_ip} remote-as 1{octets[2]}"

        for i in range(1, total_router + 1):
            bgp_commands += f"\nneighbor 1.1.{i}.1 remote-as {MY_BGP_AS_NUM}"
            bgp_commands += f"\nneighbor 1.1.{i}.1 update-source lo"

        setup_dhcp(infs_user, ip_infs_user, subnets_user)

        for subnet in subnets:
            commands += f"\nnetwork {subnet}/24 area 0"
        commands += f"\nnetwork {lo_ip}/32 area 0"

        commands += "\nexit"
        # bgp_commands += "\nredistribute ospf"
        commands += bgp_commands
        commands += "\nexit\nexit\nwrite"

        # Start vtysh in interactive mode
        process_vtysh(commands, "OSPF and BGP")
    except Exception as e:
        print(f"An error occurred: {e}")
