#include <linux/bpf.h>
#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/udp.h>

#define MY_NODE_ID 0
#define MY_OPTION_TYPE 31 // Custom option type
#define REGISTER 1
#define DROP 0
#define MAX_CHECKING 4

static inline __u16 iph_csum(struct iphdr *iph, void *data_end)
{
    __u32 sum = 0;
    __u16 *buf = (__u16 *)iph;
    __u16 ihl = iph->ihl << 2; // Convert IHL to bytes

    // Reset checksum field
    iph->check = 0;

    // Sum all 16-bit words in the header - 60 bytes is maximum size of IP header
    for (__u8 i = 0; i < ihl && i < 60; i += 2)
    {
        if ((void *)(buf + 1) > data_end)
        {
            break;
        }
        sum += *buf++;
    }

    // Add overflow (carry folding)
    for (__u8 i = 0; sum >> 16 && i < MAX_CHECKING; i += 1)
    {
        sum = (sum & 0xFFFF) + (sum >> 16);
    }

    // Take one's complement
    return ~sum;
}

// CustomIPOption structure
struct CustomIPEntry
{
    __u16 id_val;
    __u8 inter_val;
    __u8 cmd_val;
} __attribute__((packed)); // Ensure packed struct to prevent padding issues

struct FirewallRequest
{
    __u32 entry_inter_val;
    __u32 ip_dst;
    __u32 entry_cmd_val;
};

// entry.inter_val, pkt_dst, entry.cmd_val

BPF_QUEUE(queue, struct FirewallRequest, 10240);

int op_ebpf(struct xdp_md *ctx)
{
    void *data = (void *)(long)ctx->data;
    void *data_end = (void *)(long)ctx->data_end;

    struct ethhdr *eth = data;

    // Check if the Ethernet header is within bounds
    if ((void *)eth + sizeof(*eth) > data_end)
        return XDP_PASS;

    struct iphdr *ip = data + sizeof(*eth);

    // Check if the IP header is within bounds
    if ((void *)ip + sizeof(*ip) > data_end)
        return XDP_PASS;

    if (ip->version != 4)
    {
        return XDP_PASS;
    }

    // Check if the packet has IP options
    int options_len = (ip->ihl * 4) - sizeof(struct iphdr);
    __u8 *options = (__u8 *)(ip + 1);

    // bpf_trace_printk("Packet received on interface %d and option length %d\n", ctx->ingress_ifindex, options_len);

    __u8 is_register = 1;
    __u8 is_exist_custom_option = 0;
    if (options_len > 0 && (void *)(options + 4) < data_end)
    { // Ensure you're within the bounds (type + length + count = 4 bytes)
        __u8 option_type = options[0];
        if (option_type == MY_OPTION_TYPE)
        {
            is_exist_custom_option = 1;
            __u8 option_length = options[1];
            __u16 option_count = bpf_ntohs(*((__u16 *)(options + 2)));

            // if ((void *)(options + option_length) > data_end)
            // return XDP_PASS; // Prevent out-of-bounds access

            // Parse entries
            for (__u8 j = 0; j < option_count && j < MAX_CHECKING; j++)
            {
                struct CustomIPEntry *entry = (struct CustomIPEntry *)(options + 4 + (j * sizeof(struct CustomIPEntry)));

                if ((void *)(entry + 1) > data_end)
                    break; // Ensure we don't access out of bounds

                __u16 id_val = bpf_ntohs(entry->id_val);
                __u8 inter_val = entry->inter_val;
                __u8 cmd_val = entry->cmd_val;

                // bpf_trace_printk("Option %u", option_type);
                // bpf_trace_printk("Length %u", option_length);
                // bpf_trace_printk("Count %d", (__u32)option_count);
                // bpf_trace_printk("ID %u, Inter %u, CMD %u", (__u32)id_val, (__u32)inter_val, (__u32)cmd_val);

                if (cmd_val == DROP)
                {
                    if (id_val == MY_NODE_ID)
                    {
                        struct FirewallRequest req = {};
                        req.ip_dst = ip->daddr; // Store as __u32 (network byte order)
                        req.entry_inter_val = inter_val;
                        req.entry_cmd_val = DROP;
                        queue.push(&req, 0);
                    }
                }
                else if (cmd_val == REGISTER)
                {
                    is_register = 0;
                }
            }
        }
    }

    if (is_register)
    {
        if (is_exist_custom_option)
        {
            int shift_data_length = 38;             // 14 bytes ether + 20 bytes IP + 4 bytes custom option header
            int ret = bpf_xdp_adjust_head(ctx, -4); // Add 4 bytes at the head
            if (ret < 0)
            {
                return XDP_PASS; // If adjustment fails, pass the packet
            }

            // Re-fetch pointers after adjustment
            void *data = (void *)(long)ctx->data;
            void *data_end = (void *)(long)ctx->data_end;

            __u8 *data_bytes = (__u8 *)data;
            // Ensure we don't exceed the packet bounds
            if (data + shift_data_length + 4 > data_end)
            {
                return XDP_DROP; // Drop the packet if it's too small
            }

            // Shift data backward (from the end toward the beginning)
            for (int i = 0; i < shift_data_length; i++)
            {
                data_bytes[i] = data_bytes[i + 4];
            }

            struct ethhdr *eth = data;
            // Check if the Ethernet header is within bounds
            if ((void *)eth + sizeof(*eth) > data_end)
                return XDP_PASS;
            struct iphdr *ip = data + sizeof(*eth);
            // Check if the IP header is within bounds
            if ((void *)ip + sizeof(*ip) > data_end)
                return XDP_PASS;
            // Check if the packet has IP options
            int options_len = (ip->ihl * 4) - sizeof(struct iphdr);
            __u8 *options = (__u8 *)(ip + 1);

            if ((void *)(options + 4) > data_end)
                return XDP_PASS;

            options[1] += 4;
            __u16 option_count = bpf_ntohs(*((__u16 *)(options + 2)));
            option_count += 1;
            *((__u16 *)(options + 2)) = bpf_htons(option_count);

            if ((void *)(options + 4) > data_end)
                return XDP_PASS;

            // Parse entries
            struct CustomIPEntry *entry = (struct CustomIPEntry *)(options + 4);

            entry->id_val = bpf_ntohs(MY_NODE_ID);
            entry->inter_val = ctx->ingress_ifindex;
            entry->cmd_val = REGISTER;

            int new_header_size = sizeof(struct iphdr) + options_len + sizeof(struct CustomIPEntry); // 28 + 4 = 32 bytes
            ip->ihl = new_header_size / 4;
            ip->tot_len = htons(ntohs(ip->tot_len) + sizeof(struct CustomIPEntry));
            ip->check = iph_csum(ip, data_end); // Invalidate checksum
        }
        else
        {
            int shift_data_length = 34;             // 14 bytes ether + 20 bytes IP
            int ret = bpf_xdp_adjust_head(ctx, -8); // Add 8 bytes at the head
            if (ret < 0)
            {
                return XDP_PASS; // If adjustment fails, pass the packet
            }

            // Re-fetch pointers after adjustment
            void *data = (void *)(long)ctx->data;
            void *data_end = (void *)(long)ctx->data_end;

            __u8 *data_bytes = (__u8 *)data;
            // Ensure we don't exceed the packet bounds
            if (data + shift_data_length + 8 > data_end)
            {
                return XDP_DROP; // Drop the packet if it's too small
            }

            // Shift data backward (from the end toward the beginning)
            for (int i = 0; i < shift_data_length; i++)
            {
                data_bytes[i] = data_bytes[i + 8];
            }

            struct ethhdr *eth = data;
            // Check if the Ethernet header is within bounds
            if ((void *)eth + sizeof(*eth) > data_end)
                return XDP_PASS;
            struct iphdr *ip = data + sizeof(*eth);
            // Check if the IP header is within bounds
            if ((void *)ip + sizeof(*ip) > data_end)
                return XDP_PASS;
            // Check if the packet has IP options
            int options_len = (ip->ihl * 4) - sizeof(struct iphdr);
            __u8 *options = (__u8 *)(ip + 1);

            if ((void *)(options + 8) > data_end)
                return XDP_PASS;

            options[0] = MY_OPTION_TYPE;              // option type
            options[1] = 8;                           // option length
            *((__u16 *)(options + 2)) = bpf_htons(1); // option count

            // Parse entries
            struct CustomIPEntry *entry = (struct CustomIPEntry *)(options + 4);

            entry->id_val = bpf_ntohs(MY_NODE_ID);
            entry->inter_val = ctx->ingress_ifindex;
            entry->cmd_val = REGISTER;

            int new_header_size = sizeof(struct iphdr) + options_len + 4 + sizeof(struct CustomIPEntry); // 20 + 8 = 28 bytes
            ip->ihl = new_header_size / 4;
            ip->tot_len = htons(ntohs(ip->tot_len) + 4 + sizeof(struct CustomIPEntry));
            ip->check = iph_csum(ip, data_end); // Invalidate checksum
        }
    }

    return XDP_PASS;
}
