#include <stdio.h>
#include <Python.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <linux/types.h>
#include <linux/netfilter.h>
#include <netinet/ip.h>
#include <libnetfilter_queue/libnetfilter_queue.h>
#include <arpa/inet.h> // Add this header for inet_ntop
#include <string.h>    // Add this header for memcpy

#define MY_OPTION_TYPE 31
#define REGISTER 1
#define DROP 0

int node_id;
PyObject *put_func;
PyGILState_STATE gstate;

// Define your CustomIPOption and CustomIPEntry structures
typedef struct
{
    uint16_t id_val;
    uint8_t inter_val;
    uint8_t cmd_val;
} CustomIPEntry;

typedef struct
{
    uint8_t type;
    uint8_t length;
    uint16_t count;
    CustomIPEntry data[0]; // flexible array to store data entries
} CustomIPOption;

// Custom packet processing function
int process_packet(struct nfq_data *nfa, struct nfq_q_handle *qh)
{
    struct nfqnl_msg_packet_hdr *ph;
    unsigned char *data;
    int ret;
    uint32_t id, inter_num;

    // Get packet header
    ph = nfq_get_msg_packet_hdr(nfa);
    if (!ph)
    {
        fprintf(stderr, "Failed to get packet header\n");
        return 0;
    }
    id = ntohl(ph->packet_id);

    // Get packet mark (interface number)
    inter_num = nfq_get_nfmark(nfa);

    // Get packet payload
    ret = nfq_get_payload(nfa, &data);
    if (ret < 0)
    {
        fprintf(stderr, "Failed to get payload\n");
        return 0;
    }

    struct iphdr *ip_header = (struct iphdr *)data;
    if (ip_header->version != 4)
    {
        // printf("Not an IPv4 packet\n");
        nfq_set_verdict(qh, id, NF_ACCEPT, 0, NULL);
        return 0;
    }

    // char src_ip[INET_ADDRSTRLEN], dst_ip[INET_ADDRSTRLEN];
    // inet_ntop(AF_INET, &(ip_header->saddr), src_ip, INET_ADDRSTRLEN);
    // inet_ntop(AF_INET, &(ip_header->daddr), dst_ip, INET_ADDRSTRLEN);
    // printf("Processing packet: %s -> %s\n", src_ip, dst_ip);

    int is_register_me = 1;
    int option_index = -1;
    unsigned char *options = (unsigned char *)(ip_header + 1);
    int options_len = (ip_header->ihl * 4) - sizeof(struct iphdr);
    // printf("Option len: %d\n", options_len);

    // Check for custom options
    for (int i = 0; i < options_len;)
    {
        uint8_t option_type = options[i];
        uint8_t option_length = options[i + 1]; // Length of the current option
        uint16_t option_count;

        if (option_type == 1 || option_type == 0)
        {
            option_length = 1;
        }

        // printf("Option type: %d - length: %d\n", option_type, option_length);

        if (option_type == MY_OPTION_TYPE)
        {
            CustomIPOption *opt = (CustomIPOption *)&options[i];
            option_count = ntohs(opt->count);

            option_index = i;
            // Process entries in the custom option (simplified)
            for (int j = 0; j < option_count; j++)
            {
                CustomIPEntry entry = opt->data[j];
                entry.inter_val = entry.inter_val;  // Ensure correct byte order
                entry.cmd_val = entry.cmd_val;      // Ensure correct byte order
                entry.id_val = ntohs(entry.id_val); // Ensure correct byte order

                if (entry.cmd_val == DROP)
                {
                    if (entry.id_val == node_id)
                    {
                        char pkt_dst[INET_ADDRSTRLEN];
                        inet_ntop(AF_INET, &(ip_header->daddr), pkt_dst, INET_ADDRSTRLEN);

                        gstate = PyGILState_Ensure();
                        PyObject *py_tuple = Py_BuildValue("(isI)", entry.inter_val, pkt_dst, entry.cmd_val);
                        // Call queue.put((inter_num, pkt_dst, cmd))
                        PyObject_CallFunctionObjArgs(put_func, py_tuple, NULL);
                        // Clean up
                        Py_XDECREF(py_tuple);
                        // Py_XDECREF(put_func);
                        // Release GIL
                        PyGILState_Release(gstate);
                    }
                }
                else if (entry.cmd_val == REGISTER)
                {
                    is_register_me = 0;
                }
            }
            break;
        }
        i += option_length;
    }

    // Modify the packet if necessary
    if (is_register_me)
    {
        if (option_index == -1)
        {
            // printf("Adding new option\n");

            // Allocating memory for the CustomIPOption structure + space for entry_count CustomIPEntry structures
            CustomIPOption *new_opt;
            uint16_t entry_count = 1; // Number of entries you want to store
            int new_option_size = sizeof(CustomIPOption) + sizeof(CustomIPEntry) * entry_count;
            new_opt = malloc(new_option_size);

            if (!new_opt)
            {
                fprintf(stderr, "Memory allocation failed for new option\n");
                return 0;
            }

            new_opt->type = MY_OPTION_TYPE;
            new_opt->length = new_option_size; // Set length to 1 for a single entry
            new_opt->count = htons(entry_count);
            new_opt->data[0].id_val = ntohs(node_id);
            new_opt->data[0].inter_val = inter_num;
            new_opt->data[0].cmd_val = REGISTER;

            // // Print the new option for debugging
            // printf("New option constructed:\n");
            // for (int i = 0; i < new_option_size; i++)
            // {
            //     printf("%02x ", ((unsigned char *)new_opt)[i]);
            // }
            // printf("\n");

            // Update the IP header length
            int new_header_size = sizeof(struct iphdr) + options_len + new_option_size; // 20 + 8 = 28 bytes
            ip_header->ihl = new_header_size / 4;
            ip_header->tot_len = htons(ntohs(ip_header->tot_len) + new_option_size);
            ip_header->check = 0; // Invalidate checksum

            // Shift the payload to make space for the new option
            memmove(options + new_option_size, options, ret - (options - data));

            // Copy the new option into the packet
            memcpy(options, new_opt, new_option_size);

            free(new_opt);

            nfq_ip_set_checksum(ip_header);
            nfq_set_verdict(qh, id, NF_ACCEPT, ret + new_option_size, data);
        }
        else
        {
            // printf("Appending to existing option\n");

            // Locate the existing option of type MY_OPTION_TYPE
            CustomIPOption *existing_opt = (CustomIPOption *)&options[option_index];

            // Calculate the new size of the option
            int existing_option_size = existing_opt->length;
            int new_entry_size = sizeof(CustomIPEntry); // Size of one entry (4 bytes)
            int new_option_size = existing_option_size + new_entry_size;
            uint16_t existing_count = ntohs(existing_opt->count);

            // Calculate padding to align the new option to 4 bytes

            // Allocate memory for the expanded option
            CustomIPOption *new_opt = malloc(new_option_size);
            if (!new_opt)
            {
                fprintf(stderr, "Memory allocation failed for expanded option\n");
                return 0;
            }

            // printf("count: %d\n", existing_count);

            // Copy the existing option into the new memory
            memcpy(new_opt, existing_opt, existing_option_size);

            // Add the new entry
            new_opt->data[existing_count].id_val = htons(node_id); // ID value
            new_opt->data[existing_count].inter_val = inter_num;   // interface num
            new_opt->data[existing_count].cmd_val = REGISTER;      // command REGISTER

            // Update the option header
            new_opt->length = new_option_size;
            new_opt->count = htons(existing_count + 1);

            // printf("count: %d; new_count: %d\n", existing_count, ntohs(new_opt->count));

            // Print the expanded option for debugging
            // printf("Expanded option constructed:\n");
            // for (int i = 0; i < new_option_size; i++)
            // {
            //     printf("%02x ", ((unsigned char *)new_opt)[i]);
            // }
            // printf("\n");

            // Shift the payload to make space for the expanded option
            memmove(options + option_index + new_option_size, options + option_index + existing_option_size, ret - (options - data) - option_index - existing_option_size);

            // Copy the expanded option into the packet
            memcpy(options + option_index, new_opt, new_option_size);

            // Free the allocated memory
            free(new_opt);

            // Update the IP header length
            int new_header_size = sizeof(struct iphdr) + options_len + new_entry_size;
            ip_header->ihl = new_header_size / 4;
            ip_header->tot_len = htons(ntohs(ip_header->tot_len) + new_entry_size);
            ip_header->check = 0; // Invalidate checksum

            // Update the IP checksum
            nfq_ip_set_checksum(ip_header);
            nfq_set_verdict(qh, id, NF_ACCEPT, ret + new_entry_size, data);
        }
    }

    nfq_set_verdict(qh, id, NF_ACCEPT, 0, NULL);
    return 0;
}

static int cb(struct nfq_q_handle *qh, struct nfgenmsg *nfmsg, struct nfq_data *nfa, void *data)
{
    process_packet(nfa, qh);
    return 0;
}

static u_int32_t print_pkt(struct nfq_data *tb)
{
    int id = 0;
    struct nfqnl_msg_packet_hdr *ph;
    struct nfqnl_msg_packet_hw *hwph;
    u_int32_t mark, ifi;
    int ret;
    unsigned char *data;

    ph = nfq_get_msg_packet_hdr(tb);
    if (ph)
    {
        id = ntohl(ph->packet_id);
        printf("hw_protocol=0x%04x hook=%u id=%u ",
               ntohs(ph->hw_protocol), ph->hook, id);
    }

    hwph = nfq_get_packet_hw(tb);
    if (hwph)
    {
        int i, hlen = ntohs(hwph->hw_addrlen);

        printf("hw_src_addr=");
        for (i = 0; i < hlen - 1; i++)
            printf("%02x:", hwph->hw_addr[i]);
        printf("%02x ", hwph->hw_addr[hlen - 1]);
    }

    mark = nfq_get_nfmark(tb);
    if (mark)
        printf("mark=%u ", mark);

    ifi = nfq_get_indev(tb);
    if (ifi)
        printf("indev=%u ", ifi);

    ifi = nfq_get_outdev(tb);
    if (ifi)
        printf("outdev=%u ", ifi);
    ifi = nfq_get_physindev(tb);
    if (ifi)
        printf("physindev=%u ", ifi);

    ifi = nfq_get_physoutdev(tb);
    if (ifi)
        printf("physoutdev=%u ", ifi);

    ret = nfq_get_payload(tb, &data);
    if (ret >= 0)
    {
        printf("payload_len=%d ", ret);
        // processPacketData (data, ret);
    }
    fputc('\n', stdout);

    return id;
}

void start_process_from_queue(PyObject *py_queue, int my_id)
{
    node_id = my_id;
    printf("Received node_id: %d\n", node_id);

    gstate = PyGILState_Ensure();
    // Get queue.put() function
    put_func = PyObject_GetAttrString(py_queue, "put");
    if (!put_func || !PyCallable_Check(put_func))
    {
        printf("Error: Cannot find queue.put()\n");
        PyGILState_Release(gstate);
        return;
    }
    PyGILState_Release(gstate);

    struct nfq_handle *h;
    struct nfq_q_handle *qh;
    int fd;
    int rv;
    char buf[4096] __attribute__((aligned));

    // printf("opening library handle\n");
    h = nfq_open();
    if (!h)
    {
        fprintf(stderr, "error during nfq_open()\n");
        exit(1);
    }

    // printf("unbinding existing nf_queue handler for AF_INET (if any)\n");
    if (nfq_unbind_pf(h, AF_INET) < 0)
    {
        fprintf(stderr, "error during nfq_unbind_pf()\n");
        exit(1);
    }

    // printf("binding nfnetlink_queue as nf_queue handler for AF_INET\n");
    if (nfq_bind_pf(h, AF_INET) < 0)
    {
        fprintf(stderr, "error during nfq_bind_pf()\n");
        exit(1);
    }

    // printf("binding this socket to queue '1'\n");
    qh = nfq_create_queue(h, 1, &cb, NULL);
    if (!qh)
    {
        fprintf(stderr, "error during nfq_create_queue()\n");
        exit(1);
    }

    // printf("setting copy_packet mode\n");
    if (nfq_set_mode(qh, NFQNL_COPY_PACKET, 0xffff) < 0)
    {
        fprintf(stderr, "can't set packet_copy mode\n");
        exit(1);
    }

    fd = nfq_fd(h);

    while ((rv = recv(fd, buf, sizeof(buf), 0)))
    {
        // printf("pkt received\n");
        nfq_handle_packet(h, buf, rv);
    }

    printf("unbinding from queue 0\n");
    nfq_destroy_queue(qh);

    printf("closing library handle\n");
    nfq_close(h);

    exit(0);
}
