import ipaddress

from scapy.all import *
from scapy.layers.inet import IPOption
from scapy.layers.l2 import Ether

import if_utils

MY_OPTION_TYPE = 31
magic_hdr = int("0x31323334", 16)


class Cmd:
    DECISION_DROP = 0
    REGISTER = 1
    UPDATE_RULE = 2


class CustomIPOption(IPOption):
    name = "CustomIPOption"
    fields_desc = [
        ByteField("option", MY_OPTION_TYPE),  # Option type
        FieldLenField("length", 4, length_of="value", fmt="B"),  # Option length
        ShortField("count", 0),  # 2 bytes for the count field
        StrLenField("value", "", length_from=lambda pkt: pkt.length - 4),  # Array data (id, inter, cmd)
    ]

    # Custom methods to add entries (id, inter, cmd) to the value field
    def add_entry(self, id_val, inter_val, cmd_val):
        entry = struct.pack('!HBB', id_val, inter_val, cmd_val)
        self.value += entry
        self.count += 1  # Update the count field
        self.length += len(entry)

    # Custom method to read a specific entry by index
    def read_entry(self, index):
        if index < self.count:
            # Calculate the offset of the entry in the value field
            offset = index * 4  # Each entry is 4 bytes (1 short, 2 bytes)
            entry_data = self.value[offset:offset + 4]
            id_val, inter_val, cmd_val = struct.unpack('!HBB', entry_data)
            return {"node_id": id_val, "inter_num": inter_val, "cmd": cmd_val}
        else:
            raise IndexError("Entry index out of range")

    # Custom method to read all entries at once
    def read_all_entry(self):
        entries = []
        for i in range(self.count):
            entry = self.read_entry(i)
            entries.append(entry)
        return entries


def deploy_notif(ip_att):
    if ip_att:
        data = {'ip_att': ip_att, 'node_id': if_utils.my_id}
        data_js = json.dumps(data)
        HOST = "10.1.1.1"
        PORT = 12344
        client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        client.connect((HOST, PORT))
        client.sendall(data_js.encode('utf-8'))  # Send JSON data
        client.close()


e_queue_deploy_notif = if_utils.EventDrivenQueue(deploy_notif)


class FirewallOpportunistic:
    firewall_queue = queue.Queue()
    map_firewall = {}

    @staticmethod
    def is_firewall_ready(pkt):
        # Stop sniffing if firewall notification is received
        if Ether in pkt and Raw in pkt and len(pkt.layers()) == 2:
            if pkt[Ether].src == "00:00:00:00:00:00" and pkt[Ether].dst == "00:00:00:00:00:00":
                raw_data = pkt[Raw].load
                magic_hdr_pkt, cmd = struct.unpack('!IB', raw_data)
                # print(f"Found packet with magic_hdr_pkt={magic_hdr_pkt} and cmd={cmd} and my_magic_hdr={magic_hdr}")
                if magic_hdr_pkt == magic_hdr and cmd == Cmd.UPDATE_RULE:
                    # print("Firewall unikernel ready! Start applying the rule")
                    return True
        return False

    @staticmethod
    def deploy_firewall_and_rule(private_if, ip_dst, cmd):
        sniff(iface=private_if, stop_filter=FirewallOpportunistic.is_firewall_ready, store=False)
        FirewallOpportunistic.add_new_rule(private_if, ip_dst, cmd)

    @staticmethod
    def add_new_rule(private_if, ip_att, cmd):
        # Define the MAC addresses and EtherType
        dst_mac = "00:00:00:00:00:00"  # Replace with the destination MAC address
        src_mac = "00:00:00:00:00:00"  # Replace with your source MAC address
        ip_bytes = ipaddress.ip_address(ip_att).packed
        message = struct.pack('!IB4s', magic_hdr, cmd, ip_bytes)
        sendp(Ether(src=src_mac, dst=dst_mac) / Raw(load=message), iface=private_if)  # Specify the interface
        # print(f"Update rule packet sent to {private_if}")
        e_queue_deploy_notif.put(ip_att)

    @staticmethod
    def setup_firewall(tuple_data):
        inter_num, ip_dst, cmd = tuple_data

        # If a firewall in corresponding interface has not been deployed
        if inter_num not in FirewallOpportunistic.map_firewall:
            public_if = f"public{if_utils.interface_tap_id}"
            private_if = f"private{if_utils.interface_tap_id}"
            new_bridge = f"br{if_utils.interface_tap_id}"
            ens_if = if_utils.PHYSICAL_INTERFACE[inter_num]

            # Create Tap interface "public" and "private"
            if_utils.create_tuntap_interface(public_if)
            if_utils.create_tuntap_interface(private_if)

            # Start listening on the private interface to deploy the firewall rule as soon as the firewall unikernel is ready
            threading.Thread(target=FirewallOpportunistic.deploy_firewall_and_rule, args=(private_if, ip_dst, cmd)).start()

            # Create a new bridge to bridge "ens$i" and "public"
            if_utils.create_bridge(new_bridge)

            # Retrieve the old bridge to remove "ens$i"
            ingress_bridge = if_utils.map_bridge_outside[inter_num]

            # If IDS was not deployed
            if ingress_bridge is None:
                # Update queue rule
                # if_utils.delete_forward_to_queue_rule(ens_if, 1)
                # if_utils.forward_to_queue_rule(private_if, 1)

                # Move IP address from ens_if to private_if
                if inter_num in if_utils.map_inf_ip:
                    ip_address = if_utils.map_inf_ip[inter_num]
                    if_utils.delete_ip_address(ens_if, ip_address)
                    if_utils.add_ip_address(private_if, ip_address, "/24")

                # Update mark rules for private_if
                # if_utils.delete_mangle_mark_rule(ens_if, inter_num)
                # if_utils.mark_mangle_forward_rule(private_if, inter_num)
                if_utils.map_inf_inside[inter_num] = private_if
            else:
                if_utils.delete_interface_from_bridge(ingress_bridge, ens_if)
                if_utils.add_interface_to_bridge(ingress_bridge, private_if)

                # Update map inf for IDS
                inter_name_ids, bridge_ids, public_if_ids = IdsOpportunistic.map_tap_inf_for_ids[inter_num]
                IdsOpportunistic.map_tap_inf_for_ids[inter_num] = (private_if, bridge_ids, public_if_ids)

            # Add "ens$i", "public" to the new bridge
            if_utils.add_interface_to_bridge(new_bridge, ens_if)
            if_utils.add_interface_to_bridge(new_bridge, public_if)

            # Start deploying firewall unikernel. Once the unikernel says it is ready, the above listening thread will fire the rule
            threading.Thread(target=if_utils.run_firewall_solo5_hvt, args=(private_if, public_if)).start()

            # Save the current state including the firewall and rule deployment
            FirewallOpportunistic.map_firewall[inter_num] = {"rule_if": private_if, "rule_list": [(ip_dst, cmd)]}
            if_utils.map_bridge_outside[inter_num] = new_bridge
            if_utils.interface_tap_id += 1
        else:
            firewall_info = FirewallOpportunistic.map_firewall[inter_num]
            private_if = firewall_info["rule_if"]
            deployed_r_lst = firewall_info["rule_list"]

            # Currently we only have one decision in cmd (DROP) and no delete rule mecha yet so no need to check different cmd
            if (ip_dst, cmd) not in deployed_r_lst:
                # print("Firewall rule already deployed, no need to add")
                # If the rule was not added
                # else:
                FirewallOpportunistic.add_new_rule(private_if, ip_dst, Cmd.DECISION_DROP)
                deployed_r_lst.append((ip_dst, cmd))


class IdsOpportunistic:
    ids_queue = queue.Queue()
    ids_deploy_id = 0
    map_tap_inf_for_ids = {}
    ids_process = None
    is_stop = True

    @staticmethod
    def process_opportunistic_ids():
        for inter_num in if_utils.map_inf_inside.keys():
            inter_name = if_utils.map_inf_inside[inter_num]

            public_if = f"public{if_utils.interface_tap_id}"
            private_if = f"private{if_utils.interface_tap_id}"
            new_bridge = f"br{if_utils.interface_tap_id}"

            if_utils.create_tuntap_interface(public_if)
            if_utils.create_tuntap_interface(private_if)
            if_utils.create_bridge(new_bridge)

            # Add ens$i or private$k of fw$k and public_if to br of IDS
            if_utils.add_interface_to_bridge(new_bridge, inter_name)
            if_utils.add_interface_to_bridge(new_bridge, public_if)

            # If NO firewall deployed yet on this inf, update outside bridge
            if if_utils.map_bridge_outside[inter_num] is None:
                if_utils.map_bridge_outside[inter_num] = new_bridge

            # Update mark rules for private_if
            # if_utils.delete_mangle_mark_rule(inter_name, inter_num)
            # if_utils.mark_mangle_forward_rule(private_if, inter_num)
            if_utils.map_inf_inside[inter_num] = private_if

            # Update queue rule
            # if_utils.delete_forward_to_queue_rule(inter_name, 1)
            # if_utils.forward_to_queue_rule(private_if, 1)

            # Move IP address from ens$i or private$k of fw$k to private_if
            if inter_num in if_utils.map_inf_ip:
                ip_address = if_utils.map_inf_ip[inter_num]
                if_utils.delete_ip_address(inter_name, ip_address)
                if_utils.add_ip_address(private_if, ip_address, "/24")

            IdsOpportunistic.map_tap_inf_for_ids[inter_num] = (inter_name, new_bridge, public_if)

            if_utils.interface_tap_id += 1

    @staticmethod
    def process_kill_opportunistic_ids():
        for inter_num in if_utils.map_inf_inside.keys():
            next_if, bridge_ids, public_if = IdsOpportunistic.map_tap_inf_for_ids[inter_num]
            private_if = if_utils.map_inf_inside[inter_num]

            # Delete public, private and bridge created by IDS
            if_utils.delete_interface_from_bridge(bridge_ids, public_if)
            if_utils.delete_interface_from_bridge(bridge_ids, next_if)
            if_utils.delete_tuntap_interface(public_if)
            if_utils.delete_tuntap_interface(private_if)
            if_utils.delete_bridge(bridge_ids)

            # If NO firewall deployed yet on this inf, update outside bridge
            if if_utils.NAME_INF in next_if:
                if_utils.map_bridge_outside[inter_num] = None

            # Update mark rules for next_inf
            # if_utils.delete_mangle_mark_rule(private_if, inter_num)
            # if_utils.mark_mangle_forward_rule(next_if, inter_num)
            if_utils.map_inf_inside[inter_num] = next_if

            # Update queue rule
            # if_utils.delete_forward_to_queue_rule(private_if, 1)
            # if_utils.forward_to_queue_rule(next_if, 1)

            # Move IP address from private_if to next_if
            if inter_num in if_utils.map_inf_ip:
                ip_address = if_utils.map_inf_ip[inter_num]
                # if_utils.delete_ip_address(private_if, ip_address)
                if_utils.add_ip_address(next_if, ip_address, "/24")

            del (IdsOpportunistic.map_tap_inf_for_ids[inter_num])

    @staticmethod
    def start_ids():
        IdsOpportunistic.ids_process = if_utils.run_ids_solo5_hvt(IdsOpportunistic.map_tap_inf_for_ids)
        # Monitor the process
        while not IdsOpportunistic.is_stop:
            time.sleep(1)  # Avoid busy waiting

        # If the thread is stopped and process is still running, terminate it before deleting interface to avoid Errors: Device or resource busy
        if_utils.stop_ids_solo5_hvt(IdsOpportunistic.ids_process)
        IdsOpportunistic.process_kill_opportunistic_ids()
        IdsOpportunistic.ids_process = None

    @staticmethod
    def setup_ids(tuple_data):
        node_ID, time_decided, is_ids_deployed = tuple_data

        if is_ids_deployed and not IdsOpportunistic.ids_process:
            # Generate and store IDS ID to differentiate instances deployed overtime, currently we do not use it
            new_ids_deploy_id = str(node_ID) + "_" + str(time_decided)
            if IdsOpportunistic.ids_deploy_id == 0 or IdsOpportunistic.ids_deploy_id != new_ids_deploy_id:
                IdsOpportunistic.ids_deploy_id = new_ids_deploy_id

            IdsOpportunistic.is_stop = False
            IdsOpportunistic.process_opportunistic_ids()
            # print("Its time for deployment of IDS", IdsOpportunistic.ids_deploy_id)
            ids_thread = threading.Thread(target=IdsOpportunistic.start_ids)
            ids_thread.start()
        elif not is_ids_deployed:
            # set a return condition to kill IDS process and restore iptables rules
            IdsOpportunistic.is_stop = True

        # waiting 20s for the procedure finish in case the queue is filled continuously
        time.sleep(20)


def update_unikernel():
    while True:
        # if a firewall is required to deploy
        if not FirewallOpportunistic.firewall_queue.empty():
            while not FirewallOpportunistic.firewall_queue.empty():
                tuple_data = FirewallOpportunistic.firewall_queue.get()
                FirewallOpportunistic.setup_firewall(tuple_data)
        # if an IDS is required to deploy or removed
        elif not IdsOpportunistic.ids_queue.empty():
            tuple_data = ""
            while not IdsOpportunistic.ids_queue.empty():
                tuple_data = IdsOpportunistic.ids_queue.get()
            if tuple_data:
                IdsOpportunistic.setup_ids(tuple_data)
