import queue
import socket
import struct
import subprocess
import threading
import time
import json

from scapy.all import sendp, sniff, Raw, get_if_hwaddr
from scapy.layers.l2 import Ether
from scapy.layers.inet import IP

import if_utils
from if_utils import update_ospf_rule, add_ip_address, NAME_INF, map_inf_ip, disable_NetworkManager_auto_ip, drop_internet_and_allow_VC_server_rule, MIN_INTER_NUMBER, DIR, wait_for_ip
from opportunistic_node import process_opportunistic_node, start_process_from_queue, check_local_packet
from opportunistic_utils import update_unikernel, IdsOpportunistic, magic_hdr

# Configuration
LISTEN_PORT = 12346  # Specify the port number to listen on
CONVERGE_TIME_THRESHOLD = 10

update_neighbor_info_queue = queue.Queue()
client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

while True:
    try:
        client_socket.connect(('192.168.122.1', LISTEN_PORT))
        print("Connected successfully!")
        break  # Exit loop once connected
    except (socket.error, ConnectionRefusedError) as e:
        # print(f"Connection failed: {e}. Retrying in 2 seconds...")
        time.sleep(2)  # Wait before retrying


class State:
    INIT = 0
    COMPUTING = 1
    DECIDED_IN_VC = 2
    DECIDED_NOT_VC = 3


class Cmd:
    REGISTER = 1
    REPLY_REGISTER = 2
    SHARE_INFO = 3
    ACCEPT = 4
    DROP_ME = 5


class VcNode:
    def __init__(self, state, degree, round_val, myid, interface):
        self.state = state
        self.degree = degree
        self.round_val = round_val
        self.myid = myid
        self.interface = interface
        self.pointer = 0
        self.color = degree
        self.color_id = myid
        self.converge_time = 0


class NeighborNode:
    def __init__(self, neighbor_id, neighbor_degree, neighbor_color, neighbor_color_id, neighbor_pointer, neighbor_mac, iface):
        self.iface = iface
        self.mac = neighbor_mac
        self.pointer = neighbor_pointer
        self.color = neighbor_color
        self.color_id = neighbor_color_id
        self.degree = neighbor_degree
        self.id = neighbor_id


map_neighbor = {}

node = VcNode(State.INIT, 0, 0, 0, {})


# Function to broadcast an Ethernet packet
def broadcast_packet(message):
    global node

    for iface in node.interface.keys():
        src_mac = node.interface[iface]
        dst_mac = "00:00:00:00:00:00"
        packet = Ether(src=src_mac, dst=dst_mac) / Raw(load=message)  # Broadcast MAC address
        sendp(packet, iface=iface)  # Specify the interface


# Function to send a packet to a specific MAC address
def send_packet(iface, dst_mac, message):
    src_mac = get_if_hwaddr(iface)
    packet = Ether(src=src_mac, dst=dst_mac) / Raw(load=message)
    sendp(packet, iface=iface)


def send_node_info():
    global node, client_socket
    message = struct.pack('!II', node.state, node.pointer)
    client_socket.sendall(message)


# Function to receive packets at Layer 2
def process_packet(iface, my_mac):
    global node, map_neighbor

    def packet_callback(packet):
        if Ether in packet and Raw in packet and len(packet.layers()) == 2:
            # print(packet)
            if packet[Ether].dst == my_mac or (packet[Ether].src != my_mac and packet[Ether].dst == "00:00:00:00:00:00"):
                raw_data = packet[Raw].load
                # print("raw data from iface", iface, ":", raw_data)
                magic_hdr_pkt = struct.unpack('!I', raw_data[:4])[0]
                if magic_hdr_pkt == magic_hdr:
                    cmd = struct.unpack('!I', raw_data[4:8])[0]

                    if cmd == Cmd.REGISTER:
                        neighbor_id, neighbor_degree, neighbor_color, neighbor_color_id, neighbor_pointer = struct.unpack('!IIIII', raw_data[8:])
                        if neighbor_id not in map_neighbor:
                            message = struct.pack('!IIIIIII', magic_hdr, Cmd.REPLY_REGISTER, node.myid, node.degree, node.color, node.color_id, node.pointer)
                            send_packet(iface, packet[Ether].src, message)
                            # update_neighbor_info_queue.put(NeighborNode(neighbor_id, neighbor_degree, neighbor_color, neighbor_color_id, neighbor_pointer, packet[Ether].src, iface))
                    elif cmd == Cmd.REPLY_REGISTER:
                        neighbor_id, neighbor_degree, neighbor_color, neighbor_color_id, neighbor_pointer = struct.unpack('!IIIII', raw_data[8:])
                        if neighbor_id not in map_neighbor:
                            map_neighbor[neighbor_id] = NeighborNode(neighbor_id, neighbor_degree, neighbor_color, neighbor_color_id, neighbor_pointer, packet[Ether].src, iface)
                            node.degree = len(map_neighbor)
                            node.color = node.degree
                            # print(f"My degree now is: {node.degree}")
                    elif cmd == Cmd.SHARE_INFO:
                        neighbor_id, neighbor_degree, neighbor_color, neighbor_color_id, neighbor_pointer = struct.unpack('!IIIII', raw_data[8:])
                        update_neighbor_info_queue.put(NeighborNode(neighbor_id, neighbor_degree, neighbor_color, neighbor_color_id, neighbor_pointer, packet[Ether].src, iface))
        elif IP in packet:
            # This check to ensure that I can apply a firewall deployment request from my IDS unikernel
            check_local_packet(packet)
    return packet_callback
    # print(f"Received packet: {packet.summary()}")


def check_all_interfaces():
    try:
        # Run the 'ip addr' command and capture the output
        output = subprocess.check_output(['ip', 'addr'], text=True)

        # Split the output into lines
        lines = output.splitlines()
        interface_states = {}

        current_interface = None

        # Parse the output to find each interface and its state
        for line in lines:
            # Check for lines starting with a number (which indicate the interface index)
            if line and line[0].isdigit():
                # Extract the interface name (the first word in the line)
                current_interface = line.split(': ')[1].split()[0]
            # Check if the line indicates the state of the interface
            if current_interface and 'state' in line:
                state_iface = 'up' if 'state UP' in line else 'down' if 'state DOWN' in line else 'unknown'
                interface_states[current_interface] = state_iface
                current_interface = None  # Reset for the next interface

        return interface_states
    except subprocess.CalledProcessError as e:
        return f"Error executing command: {e}"


def start_sniffing(iface):
    sniff(iface=iface, prn=process_packet(iface, get_if_hwaddr(iface)), store=False)
    print(f"Sniffing {iface} ended")


def apply_rule():
    global node, map_neighbor
    low = []
    high = []
    other = []

    pointing_neighbor_id_set = []
    for neighbor_id in map_neighbor.keys():
        neighbor = map_neighbor[neighbor_id]

        if neighbor.color < node.color or (neighbor.color == node.color and neighbor.color_id < node.myid):
            low.append(neighbor)

        if neighbor.color > node.color or (neighbor.color == node.color and neighbor.color_id > node.myid):
            high.append(neighbor)

        if neighbor.pointer == node.myid:
            pointing_neighbor_id_set.append(neighbor_id)
        else:
            # Other set is all neighbors that are not Proposer
            if neighbor.pointer == 0 or (neighbor.pointer != 0 and neighbor.color_id != neighbor.id):
                other.append(neighbor)

    is_enable = False
    # Rule 1
    if node.pointer not in pointing_neighbor_id_set and node.color != node.degree:
        node.color = node.degree
        node.color_id = node.myid
        is_enable = True
        # print("Enable Rule 1")
    # Rule 2
    elif node.pointer in pointing_neighbor_id_set:
        neighbor = map_neighbor[node.pointer]
        if (node.color, node.color_id) != max((node.degree, node.myid), (neighbor.degree, neighbor.id)):
            (node.color, node.color_id) = max((node.degree, node.myid), (neighbor.degree, neighbor.id))
            is_enable = True
            # print("Enable Rule 2")

    if not is_enable:
        d_max = None
        for n_high in high:
            if n_high.pointer == node.myid:
                if not d_max or (d_max.degree < n_high.degree or (d_max.degree == n_high.degree and d_max.id < n_high.id)):
                    d_max = n_high

        # Rule 3
        if d_max and node.pointer != d_max.id:
            node.pointer = d_max.id
            node.color = d_max.degree
            node.color_id = d_max.id
            node.state = State.DECIDED_NOT_VC
            is_enable = True
            # print("Enable Rule 3")
        else:
            # Rule 4
            if node.pointer != 0:
                for n_high in high:
                    if node.pointer == n_high.id and n_high.pointer != node.myid:
                        node.pointer = 0
                        node.color = node.degree
                        node.color_id = node.myid
                        node.state = State.DECIDED_NOT_VC
                        is_enable = True
                        # print("Enable Rule 4")
                        break

            if not is_enable:
                d_min = None
                for n_low in low:
                    if not d_min or (d_min.degree > n_low.degree or (d_min.degree == n_low.degree and d_min.id > n_low.id)):
                        d_min = n_low

                # Rule 5: Only consider those are not matched to apply this rule
                if node.pointer not in pointing_neighbor_id_set and not d_max and d_min and node.pointer != d_min.id:
                    node.pointer = d_min.id
                    node.state = State.DECIDED_IN_VC
                    is_enable = True
                    # print("Enable Rule 5")
                # Rule 6: Only consider those are not Proposer to apply this rule
                elif not (node.pointer != 0 and node.color_id == node.myid and node.state == State.DECIDED_IN_VC):
                    if node.state == State.DECIDED_IN_VC:
                        is_enable = True

                        if other:
                            for n_other in other:
                                if n_other.degree < node.degree or (n_other.degree == node.degree and n_other.id < node.myid):
                                    is_enable = False
                                    break

                        # Note: Apply Rule 6 when no other set larger degree than me
                        if is_enable:
                            node.state = State.DECIDED_NOT_VC
                            # print("Enable Rule 6")

                    elif node.state == State.DECIDED_NOT_VC:
                        if other:
                            for n_other in other:
                                if n_other.degree < node.degree or (n_other.degree == node.degree and n_other.id < node.myid):
                                    node.state = State.DECIDED_IN_VC
                                    is_enable = True
                                    # print("Enable Rule 6")
                                    break
    return is_enable


def reset_me():
    node.state = State.DECIDED_NOT_VC
    node.color = node.degree
    node.color_id = node.myid
    node.pointer = 0


def print_my_info():
    print(f"{'MyID:':<10} {node.myid}")
    print(f"{'Degree:':<10} {node.degree}")
    print(f"{'State:':<10} {'IN VC' if node.state == State.DECIDED_IN_VC else 'NOT in VC'}")
    print(f"{'Color:':<10} ({node.color}, {node.color_id})")
    print(f"{'Pointer:':<10} {node.pointer}")


def update_neighbor_info():
    global map_neighbor, node, client_socket

    while True:
        try:
            is_notify_neighbor = False
            if not update_neighbor_info_queue.empty():
                while not update_neighbor_info_queue.empty():
                    neighbor = update_neighbor_info_queue.get()

                    if neighbor.id in map_neighbor:
                        old_neighbor = map_neighbor[neighbor.id]
                        if old_neighbor.degree != neighbor.degree:
                            # print("My neighbor updated its degree, reset myself")
                            reset_me()
                            is_notify_neighbor = True

                    map_neighbor[neighbor.id] = neighbor

            old_degree = node.degree
            node.degree = len(map_neighbor)

            if old_degree != node.degree:
                # print("I updated my degree, reset myself")
                reset_me()
                is_notify_neighbor = True

            if apply_rule() or is_notify_neighbor:
                send_node_info()

                for neighbor in map_neighbor.values():
                    message = struct.pack('!IIIIIII', magic_hdr, Cmd.SHARE_INFO, node.myid, node.degree, node.color, node.color_id, node.pointer)
                    send_packet(neighbor.iface, neighbor.mac, message)

                node.converge_time = time.time()
                print_my_info()
            time.sleep(1)
        except RuntimeError as e:
            print(f"Caught an error: {e}")
            print("Maybe REPLY_REGISTER received after I send SHARE_INFO message")


def update_IDS_rule(is_reaction):
    if not is_reaction:
        project_path = DIR + "/simple-IDS"
        with open(project_path + "/rules.ml", 'r') as file:
            content = file.read()
            modified_content = content.replace('action = Some ALERT;', 'action = Some NORMAL;', 1)
        with open(project_path + "/rules.ml", 'w') as file:
            file.write(modified_content)

        try:
            subprocess.run(["sudo", "-u", "node", "/usr/local/bin/opam", "exec", "dune", "build"], check=True, cwd=project_path)
        except subprocess.CalledProcessError as e:
            print("Build failed:", e)


def proceed_self_stabilizing_VC():
    global node, client_socket

    drop_internet_and_allow_VC_server_rule("ens3", LISTEN_PORT)

    stabilizing_thread = threading.Thread(target=update_neighbor_info)
    # stabilizing_thread.daemon = True
    stabilizing_thread.start()

    # Start state INITIATE
    mac_address = get_if_hwaddr("ens3")  # Example: "00:1A:2B:3C:4D:5E"
    mac_bytes = bytes.fromhex(mac_address.replace(":", ""))
    message = struct.pack('!II6s', node.state, node.pointer, mac_bytes)
    client_socket.sendall(message)

    response = client_socket.recv(8)
    received_id, json_size = struct.unpack('!II', response)
    json_data = client_socket.recv(json_size).decode()
    params = json.loads(json_data)
    data_map = params["ip_addresses"]
    is_reaction = params["is_reaction"]
    total_router = params["total_router"]

    print(f"Received ID: {received_id}")
    print(f"Received Map: {data_map}")

    update_IDS_rule(is_reaction)

    node.myid = received_id
    node.pointer = 0
    node.state = State.DECIDED_NOT_VC
    node.color_id = node.myid

    # update_netplan(data_map, "ens")
    for iface_number_str, ip in data_map.items():
        inter_num = int(iface_number_str) + MIN_INTER_NUMBER - 1
        iface_name = f'{NAME_INF}{inter_num + 1}'
        disable_NetworkManager_auto_ip(iface_name)
        add_ip_address(iface_name, ip, "/24")
        map_inf_ip[inter_num] = ip

    if_utils.my_id = node.myid
    update_ospf_rule(data_map, total_router)

    # 1. Start introduce myself to my neighbor and also get my neighbor list
    # I want to listen to all my interfaces (except ens3) even if they are "down" in dynamic topology
    for iface, state_iface in check_all_interfaces().items():
        if iface != "lo" and iface != "ens3":
            # and state_iface == "up"):
            node.interface[iface] = get_if_hwaddr(iface)
            print("Listen to interface", iface)
            sniffer_thread = threading.Thread(target=start_sniffing, args=[iface])
            # sniffer_thread.daemon = True
            sniffer_thread.start()

    # Broadcast myself to update degree
    message = struct.pack('!IIIIIII', magic_hdr, Cmd.REGISTER, node.myid, node.degree, node.color, node.color_id, node.pointer)
    turn = 3
    while turn:
        broadcast_packet(message)
        turn -= 1
        time.sleep(5)

    # Trigger a round of self-stabilisation on my neighbors
    message = struct.pack('!IIIIIII', magic_hdr, Cmd.SHARE_INFO, node.myid, node.degree, node.color, node.color_id, node.pointer)
    broadcast_packet(message)

    # My initial state is NOT in VC and start the opportunistic process for making firewall deployment
    process_opportunistic_node()
    threading.Thread(target=start_process_from_queue).start()

    # Start process requests of unikernel update
    threading.Thread(target=update_unikernel).start()

    # Checking the state change to deploy or remove the IDS unikernel
    # If I stay at VC set more than 5s -> I deploy IDS unikernel, otherwise, IDS unikernel must be destroyed
    is_ids_deployed = False
    old_time_converge = 0
    while True:
        if old_time_converge != node.converge_time:
            if not is_ids_deployed and node.state == State.DECIDED_IN_VC and (time.time() - node.converge_time) > CONVERGE_TIME_THRESHOLD:
                is_ids_deployed = True
                IdsOpportunistic.ids_queue.put((node.myid, time.time(), is_ids_deployed))
                # print("IDS Deployment request")
                old_time_converge = node.converge_time
            elif is_ids_deployed and node.state == State.DECIDED_NOT_VC and (time.time() - node.converge_time) > CONVERGE_TIME_THRESHOLD:
                is_ids_deployed = False
                IdsOpportunistic.ids_queue.put((node.myid, time.time(), is_ids_deployed))
                # print("IDS Removal request")
                old_time_converge = node.converge_time


def main():
    wait_for_ip("ens3", "192.168.122")
    proceed_self_stabilizing_VC()


if __name__ == "__main__":
    main()
