#include <linux/bpf.h>
#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/tcp.h> // For struct tcphdr

#define SUBNET_BASE __constant_htonl(0xAC103C00)
#define REPLACE_PROBABILITY 90
const __be32 ALTERNATE_IPS[] = {__constant_htonl(0x7B640202), __constant_htonl(0x7B640303)};

#define NUM_ALTERNATE_IPS (sizeof(ALTERNATE_IPS) / sizeof(ALTERNATE_IPS[0]))
BPF_ARRAY(alt_ip_index, __u32, 1); // Single-element array to store index
BPF_HASH(ip_pool, __u32, __u32, 1024);
#define SUBNET_MASK __constant_htonl(0xFFFFFF00)
#define DEST_PORT __constant_htons(445)

#define MY_OPTION_TYPE 31 // Custom option type
#define MAX_CHECKING 4
#define MAX_CSUM_WORDS 750

static inline __u16 csum_fold_helper(__u32 csum)
{
    // Add overflow (carry folding)
    for (__u8 i = 0; csum >> 16 && i < MAX_CHECKING; i += 1)
    {
        csum = (csum & 0xFFFF) + (csum >> 16);
    }

    return ~csum;
}

static __always_inline __u32 sum16(const void* data, __u32 size, const void* data_end) {
    __u32 s = 0;
    __u16 *buf = (__u16 *)data;
    for (__u32 i = 0; i < MAX_CSUM_WORDS; i++)
    {
        if (2 * i >= size)
        {
            // bpf_trace_printk("csum4: %lx", s);
            return s; /* normal exit */
        }
        if ((void *)(buf + 1) > data_end)
        {
            // bpf_trace_printk("csum5: %lx", s);
            return 0;
        }
        s += *buf++;

        if (2 * i + 3 == size)
        {
            if (data + 2 * i + 3 > data_end)
            {
                return 0; /* should be unreachable */
            }
            __u8 byte = *((__u8 *)data + 2*i + 2);
            s += byte;
            // bpf_trace_printk("csum3: %lx, bytes: %x, byte2: %x", s, byte);
            return s;
        }
    }
    return s;
}

static inline __u16 tcp_checksum(struct iphdr *ip, struct tcphdr *tcp, void *data_end)
{
    tcp->check = 0;
    __u32 csum = 0;
    __u16 tcp_len = ntohs(ip->tot_len) - (ip->ihl * 4);

    // 1. Verify entire TCP segment fits first
    if ((void *)tcp + tcp_len > data_end)
        return 0;

    // 2. Pseudo-header
    csum += (ip->saddr >> 16) + (ip->saddr & 0xFFFF); 
    csum += (ip->daddr >> 16) + (ip->daddr & 0xFFFF);
    csum += htons(6);
    csum += htons(tcp_len);
    // bpf_trace_printk("csum1: %lx", csum);

    csum += sum16(tcp, tcp_len, data_end);
    // bpf_trace_printk("csum2: %lx", csum);

    return csum_fold_helper(csum);
}

static inline __u16 iph_csum(struct iphdr *iph, void *data_end)
{
    __u32 sum = 0;
    __u16 *buf = (__u16 *)iph;
    __u16 ihl = iph->ihl << 2; // Convert IHL to bytes

    // Reset checksum field
    iph->check = 0;

    // Sum all 16-bit words in the header - 60 bytes is maximum size of IP header
    for (__u8 i = 0; i < ihl && i < 60; i += 2)
    {
        if ((void *)(buf + 1) > data_end)
        {
            return 0;
        }
        sum += *buf++;
    }

    // Take one's complement
    return csum_fold_helper(sum);
}

int inter_op_ebpf(struct xdp_md *ctx)
{
    void *data = (void *)(long)ctx->data;
    void *data_end = (void *)(long)ctx->data_end;

    struct ethhdr *eth = data;

    // Check if the Ethernet header is within bounds
    if ((void *)eth + sizeof(*eth) > data_end)
        return XDP_PASS;

    struct iphdr *ip = data + sizeof(*eth);

    // Check if the IP header is within bounds
    if ((void *)ip + sizeof(*ip) > data_end)
        return XDP_PASS;

    if (ip->version != 4)
    {
        return XDP_PASS;
    }

    // Check if the packet has IP options
    int options_len = (ip->ihl * 4) - sizeof(struct iphdr);
    __u8 *options = (__u8 *)(ip + 1);

    //    bpf_trace_printk("Packet received on interface %d and and ihl %d option length %d\n", ctx->ingress_ifindex, ip->ihl, options_len);

    __u8 is_register = 1;
    __u8 is_exist_custom_option = 0;
    if (options_len > 0 && (void *)(options + 4) < data_end)
    { // Ensure you're within the bounds (type + length + count = 4 bytes)
        __u8 option_type = options[0];
        if (option_type == MY_OPTION_TYPE)
        {
            is_exist_custom_option = 1;
            __u8 option_length = options[1];

            __u8 *data_bytes = (__u8 *)data;
            int shift_data_length = sizeof(*eth) + sizeof(struct iphdr);

            for (int i = shift_data_length - 1; i >= 0; i--) {
                char *ptr = data_bytes + option_length;
                if ((void *)ptr + i + 1 > data_end)
                    return XDP_PASS;
            
                ptr[i] = data_bytes[i];
            }

            int ret = bpf_xdp_adjust_head(ctx, option_length); // Cut option length bytes at the head
            if (ret < 0)
            {
                return XDP_PASS; // If adjustment fails, pass the packet
            }

            data = (void *)(long)ctx->data;
            data_end = (void *)(long)ctx->data_end;

            eth = data;

            // Check if the Ethernet header is within bounds
            if ((void *)eth + sizeof(*eth) > data_end)
                return XDP_DROP;

            ip = data + sizeof(*eth);

            // Check if the IP header is within bounds
            if ((void *)ip + sizeof(*ip) > data_end)
                return XDP_DROP;

            if (ip->version != 4)
            {
                return XDP_DROP;
            }

            int new_header_size = sizeof(struct iphdr); // 20 + 8 = 28 bytes
            ip->ihl = new_header_size / 4;
            ip->tot_len = htons(ntohs(ip->tot_len) - option_length);

            // bpf_trace_printk("Packet received on interface %d and option length %d\n", ctx->ingress_ifindex, options_len);
            asm volatile("" ::: "memory");
            ip->check = iph_csum(ip, data_end);
        }

        //        int old_check = ip->check;
        // ip->check = iph_csum(ip, data_end);
        //        if (old_check != ip->check){
        //            bpf_trace_printk("Old checksum %d - New checksum %d\n", old_check, ip->check);
        //        }
    }

    if (ip->protocol != 6)
        return XDP_PASS;

    struct tcphdr *tcp = (void *)ip + (ip->ihl * 4);
    if ((void *)tcp + sizeof(*tcp) > data_end)
        return XDP_PASS;

    if (tcp->source != htons(445) && tcp->dest != htons(445)) {
        return XDP_PASS;
    }

    __u32 original_src = ip->daddr;
    __u32 original_dst = ip->daddr;
    __u32 *saved_saddr = ip_pool.lookup(&ip->saddr);
    __u32 *saved_daddr = ip_pool.lookup(&ip->daddr);

    if (saved_daddr)
    {
        ip->daddr = *saved_daddr;
//         bpf_trace_printk("Change dest");
        asm volatile("" ::: "memory");
        ip->check = iph_csum(ip, data_end);
        tcp->check = tcp_checksum(ip, tcp, data_end);
    }
    else if (saved_saddr)
    {
        ip->saddr = *saved_saddr;
//         bpf_trace_printk("Change src");
        asm volatile("" ::: "memory");
        ip->check = iph_csum(ip, data_end);
        tcp->check = tcp_checksum(ip, tcp, data_end);
    }
    else if ((ip->saddr & SUBNET_MASK) == SUBNET_BASE && tcp->dest == DEST_PORT && (ip->daddr & SUBNET_MASK) != SUBNET_BASE)
    {
        __u32 alt_ip_index_key = 0;
        __u32 *index_ptr = alt_ip_index.lookup(&alt_ip_index_key);
        if (!index_ptr)
            return XDP_PASS;
        __u32 current_index = *index_ptr;

        if (current_index < NUM_ALTERNATE_IPS)
        {
            __u32 rand_num = bpf_get_prandom_u32() % 100;
            if (rand_num < REPLACE_PROBABILITY)
            {
                // Replace destination IP
                __u32 original_ip = ip->daddr;
                __u32 new_ip;
                
                for (int i = 0; i < NUM_ALTERNATE_IPS; i++) {
                    if (i == current_index) {
                        new_ip = ALTERNATE_IPS[i];  // Verifier allows (unrolled)
                    }
                }

                ip->daddr = new_ip;

                // Optional: log the replacement
                bpf_trace_printk("Replaced 0x%x with 0x%x", original_ip, ip->daddr);

                ip_pool.insert(&original_dst, &new_ip);
                ip_pool.insert(&new_ip, &original_dst);
                ip->daddr = new_ip;
                current_index++;
                alt_ip_index.update(&alt_ip_index_key, &current_index);

                ip->check = iph_csum(ip, data_end);
                tcp->check = tcp_checksum(ip, tcp, data_end);
            }
        }
    }
    return XDP_PASS;
}
