#include <stdio.h>
#include <Python.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <linux/types.h>
#include <linux/netfilter.h>
#include <netinet/ip.h>
#include <libnetfilter_queue/libnetfilter_queue.h>
#include <arpa/inet.h>  // Add this header for inet_ntop
#include <string.h>  // Add this header for memcpy

#define MY_OPTION_TYPE 31
#define REGISTER 1
#define DROP 0

int node_id;
PyObject *put_func;
PyGILState_STATE gstate;

// Define your CustomIPOption and CustomIPEntry structures
typedef struct
{
    uint16_t id_val;
    uint8_t inter_val;
    uint8_t cmd_val;
} CustomIPEntry;

typedef struct
{
    uint8_t type;
    uint8_t length;
    uint16_t count;
    CustomIPEntry data[0]; // flexible array to store data entries
} CustomIPOption;

// Custom packet processing function
int process_packet(struct nfq_data *nfa, struct nfq_q_handle *qh)
{
    struct nfqnl_msg_packet_hdr *ph;
    unsigned char *data;
    int ret;
    uint32_t id;

    // Get packet header
    ph = nfq_get_msg_packet_hdr(nfa);
    if (!ph) {
        return 0;
    }
    id = ntohl(ph->packet_id);

    // Get packet payload
    ret = nfq_get_payload(nfa, &data);
    if (ret < 0) {
        fprintf(stderr, "Failed to get payload\n");
        return 0;
    }

    struct iphdr *ip_header = (struct iphdr *)data;
    if (ip_header->version != 4) {
        // printf("Not an IPv4 packet\n");
        nfq_set_verdict(qh, id, NF_ACCEPT, 0, NULL);
        return 0;
    }

    // char src_ip[INET_ADDRSTRLEN], dst_ip[INET_ADDRSTRLEN];
    // inet_ntop(AF_INET, &(ip_header->saddr), src_ip, INET_ADDRSTRLEN);
    // inet_ntop(AF_INET, &(ip_header->daddr), dst_ip, INET_ADDRSTRLEN);
    // printf("Processing packet: %s -> %s\n", src_ip, dst_ip);

    int is_register_me = 1;
    int option_index = -1;
    unsigned char *options = (unsigned char *)(ip_header + 1);
    int options_len = (ip_header->ihl * 4) - sizeof(struct iphdr);

    // Check for custom options
    for (int i = 0; i < options_len;) {
        uint8_t option_type = options[i];
        uint8_t option_length = options[i + 1];  // Length of the current option

        if (option_type == 1 || option_type == 0){
            option_length = 1;
        }

        // Check if this is the option we're looking for (both type and length)
        if (option_type == MY_OPTION_TYPE) {
            // Found the option, now remove it by adjusting the packet
            // printf("Found option type %d with length %d in IP options. Removing it...\n", MY_OPTION_TYPE, option_length);

            // Now, shift the remaining options and data, taking into account the padding
            memmove(options + i, options + i + option_length, ret - (options + i - data) - option_length);

            // Update IP header's length after modification
            ip_header->ihl -= (option_length / 4);  // Adjust the header length in 32-bit words

            // Ensure the total length in IP header is updated
            ip_header->tot_len = htons(ntohs(ip_header->tot_len) - option_length);  // Adjust total length

            // Update the packet length after modification
            ret -= option_length;

            nfq_ip_set_checksum(ip_header);
            nfq_set_verdict(qh, id, NF_ACCEPT, ret, data);
            return 0;
        } else {
            // Move to the next option
            i += option_length;
        }
    }
    nfq_set_verdict(qh, id, NF_ACCEPT, 0, NULL);
    return 0;
}

static int cb(struct nfq_q_handle *qh, struct nfgenmsg *nfmsg, struct nfq_data *nfa, void *data)
{
    process_packet(nfa, qh);
    return 0;
}

static u_int32_t print_pkt(struct nfq_data *tb)
{
    int id = 0;
    struct nfqnl_msg_packet_hdr *ph;
    struct nfqnl_msg_packet_hw *hwph;
    u_int32_t mark, ifi;
    int ret;
    unsigned char *data;

    ph = nfq_get_msg_packet_hdr(tb);
    if (ph) {
        id = ntohl(ph->packet_id);
        printf("hw_protocol=0x%04x hook=%u id=%u ",
            ntohs(ph->hw_protocol), ph->hook, id);
    }

    hwph = nfq_get_packet_hw(tb);
    if (hwph) {
        int i, hlen = ntohs(hwph->hw_addrlen);

        printf("hw_src_addr=");
        for (i = 0; i < hlen - 1; i++)
            printf("%02x:", hwph->hw_addr[i]);
        printf("%02x ", hwph->hw_addr[hlen - 1]);
    }

    mark = nfq_get_nfmark(tb);
    if (mark)
        printf("mark=%u ", mark);

    ifi = nfq_get_indev(tb);
    if (ifi)
        printf("indev=%u ", ifi);

    ifi = nfq_get_outdev(tb);
    if (ifi)
        printf("outdev=%u ", ifi);
    ifi = nfq_get_physindev(tb);
    if (ifi)
        printf("physindev=%u ", ifi);

    ifi = nfq_get_physoutdev(tb);
    if (ifi)
        printf("physoutdev=%u ", ifi);

    ret = nfq_get_payload(tb, &data);
    if (ret >= 0) {
        printf("payload_len=%d ", ret);
        //processPacketData (data, ret);
    }
    fputc('\n', stdout);

    return id;
}

void start_process_from_queue()
{
    struct nfq_handle *h;
    struct nfq_q_handle *qh;
    int fd;
    int rv;
    char buf[4096] __attribute__((aligned));

    printf("opening library handle\n");
    h = nfq_open();
    if (!h) {
        fprintf(stderr, "error during nfq_open()\n");
        exit(1);
    }

    printf("unbinding existing nf_queue handler for AF_INET (if any)\n");
    if (nfq_unbind_pf(h, AF_INET) < 0) {
        fprintf(stderr, "error during nfq_unbind_pf()\n");
        exit(1);
    }

    printf("binding nfnetlink_queue as nf_queue handler for AF_INET\n");
    if (nfq_bind_pf(h, AF_INET) < 0) {
        fprintf(stderr, "error during nfq_bind_pf()\n");
        exit(1);
    }

    printf("binding this socket to queue '1'\n");
    qh = nfq_create_queue(h, 1, &cb, NULL);
    if (!qh) {
        fprintf(stderr, "error during nfq_create_queue()\n");
        exit(1);
    }

    printf("setting copy_packet mode\n");
    if (nfq_set_mode(qh, NFQNL_COPY_PACKET, 0xffff) < 0) {
        fprintf(stderr, "can't set packet_copy mode\n");
        exit(1);
    }

    fd = nfq_fd(h);

    while ((rv = recv(fd, buf, sizeof(buf), 0))) {
        // printf("pkt received\n");
        nfq_handle_packet(h, buf, rv);
    }

    printf("unbinding from queue 0\n");
    nfq_destroy_queue(qh);

    printf("closing library handle\n");
    nfq_close(h);

    exit(0);
}
