from bcc import BPF
from scapy.all import *
from scapy.layers.inet import UDP, IP
import re
import ipaddress
from pathlib import Path

sys.path.append(str(Path(__file__).parent.parent.resolve()))
from ubuntu_base import if_utils

DIR = "/home/node/Desktop/gns3_unikernel_testbed"
stop_sniffing = False
peer_ips = ["", ""]  # Currently only have 2 transit ASes
bgp_area = 0


def ip_to_subnet_hex(ip):
    """Convert an IP address to its subnet base (last octet set to 0) in hex format."""
    ip_obj = ipaddress.IPv4Address(ip)
    network = ipaddress.IPv4Network(f"{ip_obj}/24", strict=False)
    subnet_ip = network.network_address
    return f"0x{int(subnet_ip):08X}"


def ip_to_hex(ip):
    """Convert an IP address to hex format."""
    return f"0x{int(ipaddress.IPv4Address(ip)):08X}"


def update_ebpf_file(file_path, subnet_ip, probability, alternate_ips):
    with open(file_path, 'r') as f:
        content = f.read()

    # Update SUBNET_BASE
    subnet_hex = ip_to_subnet_hex(subnet_ip)
    content = re.sub(
        r'(#define SUBNET_BASE __constant_htonl\()0x[0-9A-Fa-f]+(\))',
        rf'\g<1>{subnet_hex}\g<2>',  # Using \g<n> for group references
        content
    )

    # Update REPLACE_PROBABILITY
    content = re.sub(
        r'(#define REPLACE_PROBABILITY )\d+',
        rf'\g<1>{probability}',  # Using \g<1> instead of \1
        content
    )

    # Update ALTERNATE_IPS
    alternate_hex = ', '.join(f'__constant_htonl({ip_to_hex(ip)})' for ip in alternate_ips)
    content = re.sub(
        r'(const __be32 ALTERNATE_IPS\[\] = \{)[^}]*(};)',
        rf'\g<1>{alternate_hex}\g<2>',
        content,
        flags=re.DOTALL
    )

    with open(file_path, 'w') as f:
        f.write(content)


def update_pref_as(peer_ip):
    global bgp_area
    commands = "configure terminal"
    commands += f"\nrouter bgp {bgp_area}"
    commands += f"\nneighbor {peer_ip} weight 500"
    commands += "\nexit\nexit\nwrite"
    if_utils.process_vtysh(commands, "preferred AS")


def packet_handler(pkt):
    global stop_sniffing, peer_ips

    if pkt.haslayer(UDP) and pkt[UDP].dport == 5000:
        # print(f"From {pkt[IP].src}: {pkt[UDP].payload}")

        udp_payload = bytes(pkt[UDP].payload)
        try:
            json_array = json.loads(udp_payload.decode('utf-8'))

            # print(f"Intercepted from {pkt[IP].src}:")
            # print(json_array)
            file_path = DIR + '/ubuntu_intermediate/inter_op_ebpf.c'
            update_ebpf_file(file_path, pkt[IP].dst, json_array['p_NAT'], json_array['handled_ips'])

            is_pref_ctrl_as = json_array['is_pref_ctrl_as']
            if is_pref_ctrl_as:
                update_pref_as(peer_ips[0])
            else:
                update_pref_as(peer_ips[1])

            stop_sniffing = True
        except (UnicodeDecodeError, json.JSONDecodeError):
            print(f"Raw data: {udp_payload.hex()}")


def update_ebpf_sniff():
    sniff(iface="ens3", filter="udp port 5000", prn=packet_handler, stop_filter=lambda _: stop_sniffing)
    print("Sniffing complete")


def start_bgp(public_ip):
    global peer_ips, bgp_area

    parts = public_ip.split('.')
    commands = "configure terminal"
    bgp_commands = f"\nrouter bgp {bgp_area}"
    bgp_index = 0

    if public_ip.startswith('123.100'):
        bgp_index = parts[2]
        bgp_commands += f"\nbgp router-id 2.2.2.{bgp_index}"
        bgp_commands += f'\nnetwork 10.0.{bgp_index}.0/24'

    parts[-1] = "254"
    peer_ip = ".".join(parts)
    if public_ip.startswith('123.100'):  # if ip of the ctrl AS
        neighbor_area = "65000"
        bgp_commands += f"\nneighbor {peer_ip} remote-as {neighbor_area}"
        peer_ips[0] = peer_ip
    else:
        neighbor_area = "65" + parts[0]
        bgp_commands += f"\nneighbor {peer_ip} remote-as {neighbor_area}"
        peer_ips[1] = peer_ip

    commands += bgp_commands
    commands += "\nexit\nexit\nwrite"
    if_utils.process_vtysh(commands, "OSPF and BGP")
    return bgp_index


def setup_uncontrolled_AS(if_index):
    while True:
        public_ip = if_utils.wait_for_ip(f"ens{if_index}", "124")
        start_bgp(public_ip)
        if_index += 1
        break  # currently we only test one uncontrolled AS


def main():
    global bgp_area
    # subprocess.run(f"sudo {DIR}/ubuntu_intermediate/./initiate.sh", shell=True, check=True)
    # print("./initiate is executed")
    public_ip = if_utils.wait_for_ip("ens3", "123.100")
    bgp_area = "1" + public_ip.split('.')[2]
    bgp_index = start_bgp(public_ip)
    # if_utils.apply_nat_rule("10.0.0.1", f"{parts[0]}.{parts[1]}.{parts[2]}.101")

    if_utils.disable_NetworkManager_auto_ip("ens4")
    if_utils.add_ip_address("ens4", f"10.0.{bgp_index}.254", "/24")
    if_utils.setup_dhcp(["ens4"], [f"10.0.{bgp_index}.254"], [f"10.0.{bgp_index}.0"])

    # Run thread to setup connection with uncontrolled AS - start at ens5
    threading.Thread(target=setup_uncontrolled_AS, args=[5]).start()

    b = BPF(src_file=f"inter_op_ebpf.c")
    fn = b.load_func("inter_op_ebpf", BPF.XDP)

    for inter_num in range(2, 13):
        inter_name = f"ens{inter_num + 1}"
        b.attach_xdp(inter_name, fn, 0)
        print("Attach the ebpf program at", inter_name)

    update_ebpf_sniff()

    for inter_num in range(2, 13):
        inter_name = f"ens{inter_num + 1}"
        print("Detach the ebpf program at", inter_name)
        b.remove_xdp(inter_name, 0)

    b = BPF(src_file=f"inter_op_ebpf.c")
    fn = b.load_func("inter_op_ebpf", BPF.XDP)

    for inter_num in range(2, 13):
        inter_name = f"ens{inter_num + 1}"
        b.attach_xdp(inter_name, fn, 0)
        print("Attach the ebpf program at", inter_name)

    try:
        b.trace_print()
    except Exception:
        print("Interrupted")

        for inter_num in range(2, 13):
            inter_name = f"ens{inter_num + 1}"
            b.remove_xdp(inter_name, 0)


if __name__ == "__main__":
    main()
